/*
 * Copyright 2003 Phil Mellor <monkeyson@users.sourceforge.net>
 * Copyright 2007 James Bursa <bursa@users.sourceforge.net>
 * Copyright 2004 Andrew Timmins <atimmins@blueyonder.co.uk>
 *
 * This file is part of NetSurf, http://www.netsurf-browser.org/
 *
 * NetSurf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * NetSurf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <locale.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <ucontext.h>
#include <execinfo.h>
#include <dlfcn.h>
#include <sys/utsname.h>
#include <libxml/encoding.h>
#include <libxml/globals.h>
#include <libxml/xmlversion.h>
#include "utils/config.h"
#include "content/fetch.h"
#include "content/fetchcache.h"
#include "content/urldb.h"
#include "desktop/netsurf.h"
#include "desktop/browser.h"
#include "desktop/gui.h"
#include "utils/log.h"
#include "utils/url.h"
#include "utils/utf8.h"
#include "utils/utils.h"

bool netsurf_quit = false;
bool netsurf_embedded = true;
bool verbose_log = true;

static void netsurf_poll(void);
static void lib_init(void);

/**
 * The regular main(), when not embedded
 */

int main(int argc, char** argv)
{
	setbuf(stderr, NULL);
	netsurf_embedded = false;
	return netsurf_main(argc, argv);
}

/** 
 * Backtrace support (handle exception signals)
 */
struct SISigCodes {
	int code;
	const char* desc; 
}; 

struct SISigCodes g_SISigCodesAny[] = {
	{ SI_USER, "SI_USER - sent by kill(2) or raise(3)"},
 	{ SI_KERNEL, "SI_KERNEL - Sent by the kernel"},
 	{ SI_QUEUE, "SI_QUEUE - sigqueue(2)"},
    { SI_TIMER, "SI_TIMER - POSIX timer expired"},
    { SI_MESGQ, "SI_MESGQ - POSIX message queue state changed (since Linux 2.6.6); see mq_notify(3)"}, 
    { SI_ASYNCIO, "AIO completed"},
	{ SI_SIGIO, "queued SIGIO"},
    { SI_TKILL, "tkill(2) or tgkill(2) (since Linux 2.4.19)"},
	{ 0, NULL }
}; 

struct SISigCodes g_SISigCodesILL[] = {
    {ILL_ILLOPC, "ILL_ILLOPC - illegal opcode"},
    {ILL_ILLOPN, "ILL_ILLOPN - illegal operand "},
    {ILL_ILLADR, "ILL_ILLADR - illegal addressing mode "},
    {ILL_ILLTRP, "ILL_ILLTRP - illegal trap"}, 
    {ILL_PRVOPC, "ILL_PRVOPC - privileged opcode "},
    {ILL_PRVREG, "ILL_PRVREG - privileged register "},
    {ILL_COPROC, "ILL_COPROC - coprocessor error "},
    {ILL_BADSTK, "ILL_BADSTK - internal stack error"},
	{ 0, NULL }
}; 
		
struct SISigCodes g_SISigCodesFPE[] = {
    {FPE_INTDIV, "FPE_INTDIV - integer divide by zero "},
    {FPE_INTOVF, "FPE_INTOVF - integer overflow "},
    {FPE_FLTDIV, "FPE_FLTDIV - floating-point divide by zero "},
    {FPE_FLTOVF, "FPE_FLTOVF - floating-point overflow "},
    {FPE_FLTUND, "FPE_FLTUND - floating-point underflow "},
    {FPE_FLTRES, "FPE_FLTRES - floating-point inexact result "},
    {FPE_FLTINV, "FPE_FLTINV - floating-point invalid operation "},
    {FPE_FLTSUB, "FPE_FLTSUB - subscript out of range"},
	{ 0, NULL }
}; 
		
struct SISigCodes g_SISigCodesSEGV[] = {
    {SEGV_MAPERR, "SEGV_MAPERR - address not mapped to object "},
    {SEGV_ACCERR, "SEGV_ACCERR - invalid permissions for mapped object"},
	{ 0, NULL }
}; 
		
struct SISigCodes g_SISigCodesBUS[] = {
    {BUS_ADRALN, "BUS_ADRALN - invalid address alignment "},
    {BUS_ADRERR, "BUS_ADRERR - nonexistent physical address "},
    {BUS_OBJERR, "BUS_OBJERR - object-specific hardware error"},
	{ 0, NULL }
}; 
		
struct SISigCodes g_SISigCodesTRAP[] = {
    {TRAP_BRKPT, "TRAP_BRKPT - process breakpoint"},
    {TRAP_TRACE, "TRAP_TRACE - process trace trap"},	
	{ 0, NULL }
}; 
		
struct SISigCodes g_SISigCodesCLD[] = {
    {CLD_EXITED, "CLD_EXITED - child has exited"},
    {CLD_KILLED, "CLD_KILLED - child was killed"},
    {CLD_DUMPED, "CLD_DUMPED - child terminated abnormally"},
    {CLD_TRAPPED, "CLD_TRAPPED - traced child has trapped"},
    {CLD_STOPPED, "CLD_STOPPED - child has stopped"},
    {CLD_CONTINUED, "CLD_CONTINUED - stopped child has continued (since Linux 2.6.9)"},
	{ 0, NULL }
}; 
				
struct SISigCodes g_SISigCodesPOLL[] = {
    {POLL_IN, "POLL_IN - data input available"},
    {POLL_OUT, "POLL_OUT - output buffers available "},
    {POLL_MSG, "POLL_MSG - input message available "},
    {POLL_ERR, "POLL_ERR - i/o error "},
    {POLL_PRI, "POLL_PRI - high priority input available "},
    {POLL_HUP, "POLL_HUP - device disconnected"},
	{ 0, NULL }
}; 
		
struct SISigInfo {
	int sig; 
	const char* desc; 
	struct SISigCodes* codes;
}; 

struct SISigInfo g_SISigInfo[] = {
    {SIGHUP, "SIGHUP - Hangup detected on controlling terminal", g_SISigCodesAny},
    {SIGINT, "SIGINT - Interrupt from keyboard", g_SISigCodesAny},
    {SIGQUIT, "SIGQUIT - Quit from keyboard", g_SISigCodesAny},
    {SIGILL, "SIGILL - Illegal Instruction", g_SISigCodesILL},
    {SIGABRT, "SIGABRT - Abort signal from abort(3)", g_SISigCodesAny},
    {SIGFPE, "SIGFPE - Floating point exception", g_SISigCodesFPE},
    {SIGKILL, "SIGKILL - Kill signal", g_SISigCodesAny},
    {SIGSEGV, "SIGSEGV - Invalid memory reference", g_SISigCodesSEGV},
    {SIGPIPE, "SIGPIPE - Broken pipe", g_SISigCodesAny},
    {SIGALRM, "SIGALRM - Timer signal from alarm(2)", g_SISigCodesAny},
    {SIGTERM, "SIGTERM - Termination signal", g_SISigCodesAny},
    {SIGUSR1, "SIGUSR1 - User-defined signal 1", g_SISigCodesAny},
    {SIGUSR2, "SIGUSR2 - User-defined signal 2", g_SISigCodesAny},
    {SIGCHLD, "SIGCHLD - Child stopped or terminated", g_SISigCodesCLD},
    {SIGCONT, "SIGCONT - Continue if stopped", g_SISigCodesAny},
    {SIGSTOP, "SIGSTOP - Stop process", g_SISigCodesAny},
    {SIGTSTP, "SIGTSTP - Stop typed at tty", g_SISigCodesAny},
    {SIGTTIN, "SIGTTIN - tty input for background process", g_SISigCodesAny},
    {SIGTTOU, "SIGTTOU - tty output for background process", g_SISigCodesAny},
    {SIGBUS, "SIGBUS - Bus error (bad memory access)", g_SISigCodesBUS},
    {SIGPOLL, "SIGPOLL - Pollable event (Sys V).", g_SISigCodesAny},
    {SIGPROF, "SIGPROF - Profiling timer expired", g_SISigCodesAny},
    {SIGSYS, "SIGSYS - Bad argument to routine", g_SISigCodesAny},
    {SIGTRAP, "SIGTRAP - Trace/breakpoint trap", g_SISigCodesTRAP},
    {SIGURG, "SIGURG - Urgent condition on socket", g_SISigCodesAny},
    {SIGVTALRM, "SIGVTALRM - Virtual alarm clock", g_SISigCodesAny},
    {SIGXCPU, "SIGXCPU - CPU time limit exceeded", g_SISigCodesAny},
    {SIGXFSZ, "SIGXFSZ - File size limit exceeded", g_SISigCodesAny},
	{0, NULL, NULL}
}; 

static void SISigDescribe(int sig, int code, const char** text_sig, const char** text_code)
{
	*text_sig = "(unknown signal)";
	*text_code = "(unknown code)"; 
	
	for(struct SISigInfo* si = g_SISigInfo; si->desc; si++) {
		if(si->sig == sig) {
			*text_sig = si->desc;
			for(struct SISigCodes* sc = si->codes; sc && sc->desc; sc++) {
				if(sc->code == code)
					*text_code = sc->desc; 
			}
		}
	}
}

static void SIBacktraceSignalHandler(int sig, siginfo_t* pinfo, void* vctx)
{
	void* ap_Backtrace[30]; 
        // make sure all logging output gets written to disk immediately now
        if(stdout) setvbuf(stdout, NULL, _IOLBF, 0); 
        int naddr = backtrace(ap_Backtrace, 30);
	printf("SIProcess: SIGNAL %d caught.\n", sig);
	if(pinfo) {
		const char* desc_sig,* desc_code;
		SISigDescribe(sig, pinfo->si_code, &desc_sig, &desc_code);
		printf("           %s; code=%d (%s)\n", desc_sig, pinfo->si_code, desc_code);
		printf("           Mem Address=%p\n", pinfo->si_addr);
	}

	if(naddr > 0) {
		printf("BACKTRACE:\n");
		backtrace_symbols_fd(ap_Backtrace, naddr, 1);
	}
        if(stdout) { fsync(1); }
	die("Fatal exception caught.");
}


struct sigaction g_sigactions[32]; 

static void SIInstallBacktrace(int sig)
{
	struct sigaction act; 
	act.sa_sigaction = &SIBacktraceSignalHandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESETHAND|SA_SIGINFO;
	sigaction(sig, &act, &(g_sigactions[sig])); 	
}

static void SIUninstallBacktrace(int sig)
{
	sigaction(sig, &g_sigactions[sig], NULL); 	
}


// Initialisiert das sanfte Herunterfahren. 
static bool SIShutdownInit(void)
{
	SIInstallBacktrace(SIGABRT);
	SIInstallBacktrace(SIGSEGV);
	SIInstallBacktrace(SIGBUS);
	SIInstallBacktrace(SIGFPE);
	
	return true; 
}

static bool SIShutdownExit(void)
{
	SIUninstallBacktrace(SIGABRT);
	SIUninstallBacktrace(SIGSEGV);
	SIUninstallBacktrace(SIGBUS);
	SIUninstallBacktrace(SIGFPE);
	
	return true; 
}


/**
 * Gui NetSurf main().
 */

int netsurf_main(int argc, char** argv)
{
	SIShutdownInit();
	jmp_buf die_jmpbuf; 
			
	die_init(&die_jmpbuf);
	
	int r = setjmp(die_jmpbuf);
	printf("setjmp() --> %d\n", r);
	if(r) {
		SIShutdownExit();
		printf("exiting because of die()\n");
		return EXIT_FAILURE; 
	}
	
	netsurf_init(argc, argv);
	
	netsurf_main_loop();

	netsurf_exit();
	SIShutdownExit();

	return EXIT_SUCCESS;
}

/** 
  * Samsung DTV main()
  */
int Game_Main(const char *path, const char *udn)
{
	char buf[1000]; 
	sprintf(buf, "%snsfb-sdl", path);
	char* dummy[1] = {buf}; 
	chdir(path);
	
	FILE* log_stdout = freopen("./output.log", "wt", stdout);
	//if(log_stdout) setvbuf(log_stdout, NULL, _IOLBF, 0); 
	fprintf(log_stdout, "Hello Netsurf!\n");
	FILE* old_stderr = stderr; 
	if(log_stdout) stderr = log_stdout; 
		   
	int r = 0; 
	r = netsurf_main(1, dummy);
	
	stderr = old_stderr; 
	if(log_stdout) fclose(log_stdout);
	
	return r; 
}



/**
 * Gui NetSurf main loop.
 */

int netsurf_main_loop(void)
{
	while (!netsurf_quit)
		netsurf_poll();

	return 0;
}


/**
 * Initialise components used by gui NetSurf.
 */

void netsurf_init(int argc, char** argv)
{
	struct utsname utsname;

	/* Ignore SIGPIPE - this is necessary as OpenSSL can generate these
	 * and the default action is to terminate the app. There's no easy
	 * way of determining the cause of the SIGPIPE (other than using
	 * sigaction() and some mechanism for getting the file descriptor
	 * out of libcurl). However, we expect nothing else to generate a
	 * SIGPIPE, anyway, so may as well just ignore them all. */
	signal(SIGPIPE, SIG_IGN);
/*
#if !((defined(__SVR4) && defined(__sun)) || defined(__NetBSD__) || \
	defined(__OpenBSD__)) 
	stdout = stderr;
#endif
*/
	if ((argc > 1) && (argv[1][0] == '-') && (argv[1][1] == 'v') && (argv[1][2] == 0)) {
	    int argcmv;
	    verbose_log = true;
	    for (argcmv = 2; argcmv < argc; argcmv++) {
		argv[argcmv - 1] = argv[argcmv];
	    }
	    argc--;
	}

#ifdef _MEMDEBUG_H_
	memdebug_memdebug("memdump");
#endif

	LOG(("version '%s'", netsurf_version));
	if (uname(&utsname) < 0)
		LOG(("Failed to extract machine information"));
	else
		LOG(("NetSurf on <%s>, node <%s>, release <%s>, version <%s>, "
				"machine <%s>", utsname.sysname,
				utsname.nodename, utsname.release,
				utsname.version, utsname.machine));

	lib_init();
	url_init();
	gui_init(argc, argv);
	setlocale(LC_ALL, "C");
	fetch_init();
	fetchcache_init();
	gui_init2(argc, argv);
}

/**
 * Poll components which require it.
 */

void netsurf_poll(void)
{
	static unsigned int last_clean = 0;
	unsigned int current_time = wallclock();

	/* avoid calling content_clean() more often than once every 5
	 * seconds.
	 */
	if (last_clean + 500 < current_time) {
		last_clean = current_time;
		content_clean();
	}
	gui_poll(fetch_active);
	fetch_poll();
}


/**
 * Clean up components used by gui NetSurf.
 */

void netsurf_exit(void)
{
	LOG(("Closing GUI"));
	gui_quit();
	LOG(("Closing content"));
	content_quit();
	LOG(("Closing fetches"));
	fetch_quit();
	LOG(("Closing utf8"));
	utf8_finalise();
	LOG(("Destroying URLdb"));
	urldb_destroy();
	LOG(("Exited successfully"));
}


/**
 * Initialises the libraries used in NetSurf.
 */
void lib_init(void)
{
	LOG(("xmlParserVersion %s, LIBXML_VERSION_STRING %s",
			xmlParserVersion, LIBXML_VERSION_STRING));

	/* Using encoding "X-SJIS" (unknown to libxmp2/iconv) instead as
	 * "Shift-JIS" is rather popular.
	 */
	if (xmlAddEncodingAlias(xmlGetCharEncodingName(
			XML_CHAR_ENCODING_SHIFT_JIS), "X-SJIS") != 0)
		die("Failed to add encoding alias");
}
