/*
 * Copyright 2010 Florian Fischer <florianfischer@gmx.de>
 *
 * This file is part of NetSurf, http://www.netsurf-browser.org/
 *
 * NetSurf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * NetSurf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>

#include "desktop/options.h"
#include "utils/log.h"
#include "utils/utf8.h"
#include "utils/messages.h"
#include "utils/utils.h"

#include "framebuffer/fb_bookmarks.h"

static char *fb_bookmarks_file = NULL;
static bool  fb_bookmarks_changed = false;
static fb_bookmark_group_t *fb_bookmarks = NULL; 

static fb_bookmark_t* fb_bookmark_new(const char*url, const char*title)
{
        fb_bookmark_t* ret = calloc(1,sizeof(fb_bookmark_t));
        if(url) ret->url = strdup(url);
        if(title) ret->title = strdup(title);
        return ret; 
}

static void fb_bookmark_free(fb_bookmark_t* bk)
{
        if(!bk) return; 
        
        if(bk->url) {
                free(bk->url);
                bk->url = NULL;
        }
        if(bk->title) {
                free(bk->title);
                bk->title = NULL; 
        }
        free(bk);
}

static fb_bookmark_group_t* fb_bookmark_group_new(const char*name)
{
        fb_bookmark_group_t* ret = calloc(1,sizeof(fb_bookmark_group_t));
        if(name) ret->name = strdup(name);
        return ret; 
}

static void fb_bookmark_group_free(fb_bookmark_group_t* g)
{
        if(!g) return; 
        
        if(g->name)
                free(g->name);
        g->name = NULL;
        free(g);
}




void bookmarks_read(const char *file)
{
        if(fb_bookmarks_file) free(fb_bookmarks_file);
        fb_bookmarks_file = strdup(file);
        
        FILE* f = NULL; 
        char linebuf[5000] = {0}; 
        char* p = NULL; 
        bool expect_url = true; 
        fb_bookmark_group_t* current_group = NULL;
        char urlbuf[5000] = {0}; 
        
        f = fopen(fb_bookmarks_file, "rt");
        if(!f) {
                LOG(("can't read from %s", fb_bookmarks_file));
                goto done;
        }
        while((p = fgets(linebuf, sizeof(linebuf), f))) {
                // trim from right
                char* pright = p + strlen(p);
                while(pright > p && isspace(pright[-1])) {
                        pright--; *pright = 0;
                }
                // trim from left
                while(*p && isspace(*p))
                        p++;
                
                // ignore empty lines
                if(!*p) continue; 
                
                if(expect_url) {
                        if(*p == '[') {
                                // start new group
                                // cut '[' and ']' 
                                p++; pright[-1] = 0; 
                                current_group = bookmarks_add_group(p);
                                // enter the group directly, don't switch parsing state
                                continue; 
                        }
                        // start new bkmk
                        strcpy(urlbuf, p);
                }
                else {
                        // read title
                        char* title = p; 
                        // add bkmk to grp
                        if(!current_group) 
                                current_group = bookmarks_add_group("Unsorted");
                        bookmarks_add(current_group, urlbuf, title); 
                }
                // switch parsing state
                expect_url = !expect_url; 
        }
        
        // Loaded correctly. Memory version matches the file contents. 
        LOG(("loaded correctly from %s", fb_bookmarks_file));
        fb_bookmarks_changed = false; 
        
done:
        if(!bookmarks_get())
                  bookmarks_add_group("Unsorted");
        if(f) {
                fclose(f);
                f = NULL;
        }
}

/* Access to the bookmarks. Use for reading and display only! */
fb_bookmark_group_t *bookmarks_get(void)
{
        return fb_bookmarks;
}
/* Add a new bookmark group. Always at the end!! */
fb_bookmark_group_t *bookmarks_add_group(const char *name)
{
        fb_bookmark_group_t* ret = fb_bookmark_group_new(name);
        if(!fb_bookmarks)
                fb_bookmarks = ret; 
        else {
                fb_bookmark_group_t* g = fb_bookmarks; 
                while(g->next) g = g->next;
                g->next = ret; 
        }
        
        fb_bookmarks_changed = true; 
        
        return ret; 
}
/* Add a new bookmark to the given group. */
fb_bookmark_t *bookmarks_add(fb_bookmark_group_t *group, const char *url, const char *title)
{
        if(!group) return NULL; 
        
        fb_bookmark_t* ret = fb_bookmark_new(url, title);
        ret->group = group; 
        /* add to group */
        if(!group->entry)
                group->entry = ret; 
        else {
                fb_bookmark_t* b = group->entry;
                while(b->next) b = b->next;
                b->next = ret;
        }
        
        fb_bookmarks_changed = true; 
        
        return ret; 
}
/* Delete a bookmark. */
void bookmarks_delete(fb_bookmark_t *entry)
{
        if(!entry) return; 
        
        fb_bookmark_group_t* g = entry->group;
        if(g) {
                if(g->entry == entry) 
                        g->entry = entry->next;
                else {
                        fb_bookmark_t* prev = g->entry; 
                        while(prev->next && prev->next != entry) prev = prev->next; 
                        prev->next = entry->next; 
                }
        }
        fb_bookmark_free(entry);

        fb_bookmarks_changed = true; 
}
/* Update a bookmark with new information. */
void bookmarks_update(fb_bookmark_t *entry, const char *url, const char *title)
{
        if(!entry) return; 
        
        if(entry->url) free(entry->url);
        entry->url = strdup(url);
        if(entry->title) free(entry->title);
        entry->title = strdup(title);
        
        fb_bookmarks_changed = true; 
}

/* Have the bookmarks been modified thru the above functions? */
bool bookmarks_changed(void)
{
        return fb_bookmarks_changed;
}
/* Saves the bookmarks to the file from which they were loaded. 
 * Does nothing if the bookmarks have not been changed. */
void bookmarks_save(void)
{
        if(!fb_bookmarks_changed) return; // nothing to do
        
        if(!fb_bookmarks_file) {
                LOG(("fb_bookmarks_file is not set"));
                return; 
        }
        
        // save
        FILE* f = fopen(fb_bookmarks_file, "wt");
        if(!f) {
                LOG(("can't write to %s", fb_bookmarks_file));
                return; 
        }
        
        for(fb_bookmark_group_t* g = bookmarks_get(); g; g = g->next) {
                fprintf(f, "\n[%s]\n", g->name);
                
                for(fb_bookmark_t* b = g->entry; b; b = b->next) {
                        fprintf(f, "%s\n  %s\n", b->url, b->title);
                }
        }
        if(ferror(f)) {
                LOG(("error saving to file %s", fb_bookmarks_file));
        }
        fclose(f);
        
        // Saved correctly. 
        LOG(("saved to %s correctly.", fb_bookmarks_file));
        fb_bookmarks_changed = false; 
}

/* Frees the loaded bookmarks. */
void bookmarks_free(void)
{
        /* recursivly free bookmarks */
        fb_bookmark_group_t* gnext; 
        for(fb_bookmark_group_t* g = bookmarks_get(); g; g = gnext)
        {
                gnext = g->next;
                
                fb_bookmark_t* bnext;
                for(fb_bookmark_t* b = g->entry; b; b = bnext) 
                {
                        bnext = b->next; 
                        
                        fb_bookmark_free(b);
                }
                
                fb_bookmark_group_free(g);
        }
        fb_bookmarks = NULL; 
        
        /* free other stuff */
        if(fb_bookmarks_file) 
                free(fb_bookmarks_file);
        fb_bookmarks_file = NULL; 
}





