/*
 * Copyright 2008 Vincent Sanders <vince@simtec.co.uk>
 *
 * This file is part of NetSurf, http://www.netsurf-browser.org/
 *
 * NetSurf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * NetSurf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <wctype.h>

#include <SDL/SDL.h>

#include "css/css.h"
#include "desktop/gui.h"
#include "desktop/options.h"
#include "utils/messages.h"
#include "utils/utils.h"
#include "desktop/history_core.h"
#include "desktop/textinput.h"

#include "framebuffer/fb_gui.h"
#include "framebuffer/fb_tk.h"
#include "framebuffer/fb_plotters.h"
#include "framebuffer/fb_frontend.h"
#include "framebuffer/fb_cursor.h"
#include "framebuffer/fb_options.h"

#include "utils/log.h"

static SDL_Surface *sdl_screen;
static framebuffer_t* sdl_fb;

// ------------------------- mouse motion stuff --------------------------

static struct {
        uint32_t last_update_ticks;
        int32_t x;
        int32_t y; 
} g_mouse_data = { 0, 0, 0 }; 

static void 
process_mouse_motion(fbtk_widget_t* root)
{
        // send event
        fbtk_move_pointer(root, g_mouse_data.x, g_mouse_data.y, false);
        // reset data
        g_mouse_data.last_update_ticks = SDL_GetTicks();
}

// ------------------------- VKB stuff --------------------------

static SDL_TimerID fb_vkb_timer_id = NULL;
static uint32_t fb_vkb_timer_func(uint32_t interval, void* param)
{
 
        SDL_Event user_event;

        user_event.type=SDL_USEREVENT;
        user_event.user.code=2;
        user_event.user.data1=NULL;
        user_event.user.data2=NULL;
        
        SDL_PushEvent(&user_event);
        return interval; 
}

// ------------------------- --------------------------

static void
set_palette(framebuffer_t *fb)
{
        SDL_Color colors[256];
        int loop;
        for(loop=0; loop < 256; loop++){
                colors[loop].r = loop;
                colors[loop].g = loop;
                colors[loop].b = loop;
                fb->palette[loop] = loop << 16 | loop << 8 | loop;
        }

        /* Set palette */
        SDL_SetColors(sdl_screen, colors, 0, 256);

}

framebuffer_t *fb_os_init(int argc, char** argv)
{
        framebuffer_t *newfb;
        int fb_width;
        int fb_height;
        int fb_depth;


        if ((option_window_width != 0) && (option_window_height != 0)) {
                fb_width = option_window_width;
                fb_height = option_window_height;                
        } else {
                fb_width = 800;
                fb_height = 600;
        }

        fb_depth = option_fb_depth;
        if ((fb_depth != 32) && (fb_depth != 16) && (fb_depth != 8))
                fb_depth = 16; /* sanity checked depth in bpp */

        newfb = calloc(1, sizeof(framebuffer_t));
        if (newfb == NULL)
                return NULL;

        if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_EVENTTHREAD|SDL_INIT_TIMER) < 0 ) {
                fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
                return NULL;
        }
        //atexit(SDL_Quit);

        newfb->width = fb_width;
        newfb->height = fb_height;
        newfb->bpp = fb_depth;

        sdl_screen = SDL_SetVideoMode(newfb->width, 
                                      newfb->height, 
                                      newfb->bpp, 
                                      /*SDL_SWSURFACE*/
                                      SDL_HWSURFACE);

        if ( sdl_screen == NULL ) {
                fprintf(stderr, 
                        "Unable to set video: %s\n", SDL_GetError());
                free(newfb);
                return NULL;
        }
        LOG(("screen created. locking needed: %d.\n", SDL_MUSTLOCK(sdl_screen)));

        if (newfb->bpp == 8)
                set_palette(newfb);

        newfb->ptr = sdl_screen->pixels;
        newfb->linelen = sdl_screen->pitch;
        
        if(option_fb_key_repeat) {
                int delay = max(100, option_fb_key_repeat_delay);
                int interval = max(20, option_fb_key_repeat_interval); 
                SDL_EnableKeyRepeat(delay, interval);
        }
        else {
                SDL_EnableKeyRepeat(0, 0); // disable key repeat
        }
        
        SDL_EnableUNICODE(true);
        
        SDL_ShowCursor(SDL_DISABLE);
        fb_os_warp_pointer(newfb->width/2, newfb->height/2);
        
        sdl_fb = newfb; 

        return newfb;
}

void fb_os_quit(framebuffer_t *fb)
{
    LOG(("fb_os_quit"));
    SDL_Quit();
}

void fb_os_input(fbtk_widget_t *root, bool active) 
{
        int got_event;
        SDL_Event event;
        int nskey = 0;
        
        //LOG(("active=%u", active));        

        if (active)
                got_event = SDL_PollEvent(&event);
        else
                got_event = SDL_WaitEvent(&event);

        /* Do nothing if there was no event */
        if (got_event == 0) {
            //LOG(("no event"));        
            return;
        }

        //LOG(("evtype=%d", event.type));        
        switch (event.type) {
        default:
                return;
                
        case SDL_USEREVENT: 
                // currently only used for VKB timeout notification
                SDL_RemoveTimer(fb_vkb_timer_id); // always remove timer after it has fired once
                nskey = fb_vkb_timeout(window_list); // get key to input
                if(nskey) fbtk_input(root, nskey); 
                break; 
                
        case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                case SDLK_PAGEDOWN:
                        nskey = KEY_PAGE_DOWN;
                        break;
                    
                case SDLK_PAGEUP:
                        nskey = KEY_PAGE_UP;
                        break;

                case SDLK_LEFT:
                        nskey = KEY_LEFT;
                        break;

                case SDLK_RIGHT:
                        nskey = KEY_RIGHT;
                        break;

                case SDLK_DOWN:
                        nskey = KEY_DOWN;
                        break;

                case SDLK_UP:
                        nskey = KEY_UP;
                        break;
                
                case SDLK_BACKSPACE:
                        nskey = '\b'; 
                        break;
                
                case SDLK_DELETE:
                        nskey = KEY_DELETE_RIGHT;
                        break; 
                
                case SDLK_TAB:
                        nskey = KEY_TAB;
                        break; 
                
                // FF -> addtional keys for remote control usage
                case SDLK_ESCAPE:
                        nskey = KEY_ESCAPE;
                        netsurf_quit = true;
                        break;
                        
                /*case SDLK_z:
                case SDLK_RETURN:
                {
                    int px, py; 
                    fb_os_get_pointer_pos(&px, &py);
                    LOG(("RETURN @pos: %d,%d", px, py));
                    fbtk_click(root, BROWSER_MOUSE_CLICK_1);
                    break;
                }*/
                        
                case SDLK_RETURN:
                case SDLK_KP_ENTER:
                    nskey = '\r'; 
                    break;
                        
                case SDLK_HOME:
                case SDLK_LALT:
                case SDLK_RALT:
                    nskey = FB_KEY_MENU; 
                    break;
                    
                case SDLK_F1:
                    if(fb_is_writing_text(window_list)) 
                        nskey = '\b'; 
                    else
                        nskey = FB_KEY_BACK;
                    break;
                    
                case SDLK_F2:
                    nskey = FB_KEY_NAVMODE;
                    break;
                    
                case 0x140: 
                        /* this key is sent by all keys belonging to critical TV functions
                           the application must terminate in response to this key
                           otherwise the watchdog will fire and reboot the TV. */
                        nskey = KEY_ESCAPE;
                        netsurf_quit = true;
                        break;
                        
                default:
                        // FF correction: do NOT just pass on any keysyms!
                        // we want real characters!
                        nskey = event.key.keysym.unicode; 
                        // FF bugfix: SamyGO SDL doesn't do unicode transform
                        if(!nskey && event.key.keysym.sym < 0x100)
                                nskey = event.key.keysym.sym; 
                        if(iswcntrl(nskey)) {
                                LOG(("keydown key=%d(U+%04x)  non-printable", event.key.keysym.sym, event.key.keysym.unicode));
                                nskey = 0;
                        }
                        if(window_list && window_list->vkb && fb_is_writing_text(window_list)) {
                                // pass char to VKB
                                nskey = fb_vkb_input(window_list, nskey);
                                // do not enter the key just now, only after a timeout
                                SDL_RemoveTimer(fb_vkb_timer_id); // remove currently running timer, if any
                                fb_vkb_timer_id = SDL_AddTimer(window_list->vkb_state.timeout_ms, fb_vkb_timer_func, NULL);
                        }
                        break;
                }
                if(nskey) 
                        fbtk_input(root, nskey);

                break;

        case SDL_MOUSEMOTION:
                g_mouse_data.x = event.motion.x;
                g_mouse_data.y = event.motion.y; 
                if(SDL_GetTicks() >= g_mouse_data.last_update_ticks + option_fb_mouse_delay)
                        process_mouse_motion(root);
                break;

        case SDL_MOUSEBUTTONDOWN:
                // in any case update mouse pos. now
                process_mouse_motion(root);
                LOG(("btndown %d", event.button.button)); 

                switch (event.button.button) {

                case SDL_BUTTON_LEFT:
                        fbtk_click(root, BROWSER_MOUSE_PRESS_1);
                        break;

                case SDL_BUTTON_RIGHT:
                        fbtk_click(root, BROWSER_MOUSE_PRESS_2);
                        break;

                case SDL_BUTTON_WHEELUP:
                        fbtk_input(root, KEY_UP);
                        break;

                case SDL_BUTTON_WHEELDOWN:
                        fbtk_input(root, KEY_DOWN);
                        break;

                }
                break;

        case SDL_MOUSEBUTTONUP:
                LOG(("btnup %d", event.button.button)); 
                switch (event.button.button) {

                case SDL_BUTTON_LEFT:
                        fbtk_click(root, BROWSER_MOUSE_CLICK_1);
                        break;

                case SDL_BUTTON_RIGHT:
                        fbtk_click(root, BROWSER_MOUSE_CLICK_2);
                        break;

                }
                break;

        case SDL_QUIT:
                netsurf_quit = true;
                break;
        }

        if(window_list && window_list->bw && window_list->bw->window && window_list->bw->window->c_h)
        {
//			struct gui_window tgw;
//			tgw.shared = gwin;
                gui_window_place_caret(window_list->bw->window,window_list->bw->window->c_x,window_list->bw->window->c_y,window_list->bw->window->c_h);
        }
}

void
fb_os_option_override(void)
{
}

void
fb_os_lock(void)
{
    if(sdl_screen && SDL_MUSTLOCK(sdl_screen)) {
        if(SDL_LockSurface(sdl_screen) == -1)
            LOG(("error in SDL_LockSurface"));
        //LOG(("fb_os_lock"));
        if(sdl_fb) {
            //LOG(("set sdl_fb->ptr=%p", sdl_screen->pixels));
            sdl_fb->ptr = sdl_screen->pixels;
            sdl_fb->linelen = sdl_screen->pitch;
        }
    }
}

void
fb_os_unlock(void)
{
    if(sdl_screen && SDL_MUSTLOCK(sdl_screen)) {
        SDL_UnlockSurface(sdl_screen);
        //LOG(("fb_os_unlock"));
    }
}

/* called by generic code to inform os code of screen update */
void
fb_os_redraw(struct bbox_s *box)
{
        /*LOG(("%d,%d-%d,%d %d,%d", box->x0, box->y0, 
             box->x1, box->y1 , 
             box->x1 - box->x0, box->y1 - box->y0));*/
        /*
        if ((box->y1 - box->y0) < 0) {
                LOG(("WTF happened"));
                return;
        }
        */
        /*SDL_UpdateRect(sdl_screen, 
                       box->x0, 
                       box->y0, 
                       box->x1 - box->x0, 
                       box->y1 - box->y0);*/
        SDL_UpdateRect(sdl_screen, 0,0,0,0);
}

uint32_t
fb_map_rgb_32(uint8_t r, uint8_t g, uint8_t b)
{
    return SDL_MapRGB(sdl_screen->format, r, g, b);
}

void 
fb_os_get_pointer_pos(int* x, int* y)
{
    SDL_GetMouseState(x,y);
}

void 
fb_os_warp_pointer(int x, int y)
{
    SDL_WarpMouse(x, y);
}


/*
 * Local Variables:
 * c-basic-offset:8
 * End:
 */

