/*
 * Copyright 2008 Vincent Sanders <vince@simtec.co.uk>
 *
 * This file is part of NetSurf, http://www.netsurf-browser.org/
 *
 * NetSurf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * NetSurf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include <wctype.h>

#include <hubbub/hubbub.h>

#include "desktop/gui.h"
#include "desktop/plotters.h"
#include "desktop/netsurf.h"
#include "desktop/options.h"
#include "utils/log.h"
#include "utils/utf8.h"
#include "utils/messages.h"
#include "utils/utils.h"
#include "desktop/textinput.h"

#include "framebuffer/fb_gui.h"
#include "framebuffer/fb_tk.h"
#include "framebuffer/fb_bitmap.h"
#include "framebuffer/fb_frontend.h"
#include "framebuffer/fb_plotters.h"
#include "framebuffer/fb_schedule.h"
#include "framebuffer/fb_cursor.h"
#include "framebuffer/fb_findfile.h"
#include "framebuffer/fb_image_data.h"
#include "framebuffer/fb_font.h"
#include "framebuffer/fb_options.h"
#include "framebuffer/fb_bookmarks.h"

#include "content/urldb.h"
#include "desktop/history_core.h"
#include "content/fetch.h"
#include "render/box.h"

char *default_stylesheet_url;
char *adblock_stylesheet_url;
char *options_file_location;

fbtk_widget_t *fbtk;

enum fb_nav_mode fb_nav_mode = FB_NAV_LINK;

struct gui_window *input_window = NULL;
struct gui_window *search_current_window;
struct gui_window *window_list = NULL;

bool redraws_pending = false;

framebuffer_t *framebuffer;

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

/* private data for browser user widget */
struct browser_widget_s {
        struct browser_window *bw; /**< The browser window connected to this gui window */
        int scrollx, scrolly; /**< scroll offsets. */

        /* Pending window redraw state. */
        bool redraw_required; /**< flag indicating the foreground loop
                               * needs to redraw the browser widget.
                               */
        bbox_t redraw_box; /**< Area requiring redraw. */
        bool pan_required; /**< flag indicating the foreground loop
                               * needs to pan the window.
                               */
        int panx, pany; /**< Panning required. */
};


/* queue a redraw operation, co-ordinates are relative to the window */
static void
fb_queue_redraw(struct fbtk_widget_s *widget, int x0, int y0, int x1, int y1)
{
        struct browser_widget_s *bwidget = fbtk_get_userpw(widget);

        bwidget->redraw_box.x0 = MIN(bwidget->redraw_box.x0, x0);
        bwidget->redraw_box.y0 = MIN(bwidget->redraw_box.y0, y0);
        bwidget->redraw_box.x1 = MAX(bwidget->redraw_box.x1, x1);
        bwidget->redraw_box.y1 = MAX(bwidget->redraw_box.y1, y1);

        bwidget->redraw_required = true;

        fbtk_request_redraw(widget);
}

static void fb_pan(fbtk_widget_t *widget,
                   struct browser_widget_s *bwidget,
                   struct browser_window *bw)
{
        struct content *c;
        int x;
        int y;
        int width;
        int height;

        c = bw->current_content;

        if ((!c) || (c->locked))
                return;

        height = fbtk_get_height(widget);
        width = fbtk_get_width(widget);
        x = fbtk_get_x(widget);
        y = fbtk_get_y(widget);

        /* dont pan off the top */
        if ((bwidget->scrolly + bwidget->pany) < 0)
                bwidget->pany = - bwidget->scrolly;

        /* do not pan off the bottom of the content */
        if ((bwidget->scrolly + bwidget->pany) > (c->height - height))
                bwidget->pany = (c->height - height) - bwidget->scrolly;

        /* dont pan off the left */
        if ((bwidget->scrollx + bwidget->panx) < 0)
                bwidget->panx = - bwidget->scrollx;

        /* do not pan off the right of the content */
        if ((bwidget->scrollx + bwidget->panx) > (c->width - width))
                bwidget->panx = (c->width - width) - bwidget->scrollx;

        LOG(("panning %d, %d",bwidget->panx, bwidget->pany));

        /* pump up the volume. dance, dance! lets do it */
        if (bwidget->pany < 0) {
                /* we cannot pan more than a window height at a time */
                if (bwidget->pany < -height)
                        bwidget->pany = -height;

                LOG(("panning up %d", bwidget->pany));

                fb_plotters_move_block(x, y,
                                       width, height + bwidget->pany,
                                       x, y - bwidget->pany);
                bwidget->scrolly += bwidget->pany;
                fb_queue_redraw(widget, 0, 0, width, - bwidget->pany);
        }

        if (bwidget->pany > 0) {
                /* we cannot pan more than a window height at a time */
                if (bwidget->pany > height)
                        bwidget->pany = height;

                LOG(("panning down %d", bwidget->pany));

                fb_plotters_move_block(x, y + bwidget->pany,
                                       width, height - bwidget->pany,
                                       x, y);
                bwidget->scrolly += bwidget->pany;
                fb_queue_redraw(widget, 0, height - bwidget->pany, width, height);
        }

        if (bwidget->panx < 0) {
                /* we cannot pan more than a window width at a time */
                if (bwidget->panx < -width)
                        bwidget->panx = -width;

                LOG(("panning left %d", bwidget->panx));

                fb_plotters_move_block(x, y,
                                       width + bwidget->panx, height ,
                                       x - bwidget->panx, y );
                bwidget->scrollx += bwidget->panx;
                fb_queue_redraw(widget, 0, 0, -bwidget->panx, height);
        }

        if (bwidget->panx > 0) {
                /* we cannot pan more than a window width at a time */
                if (bwidget->panx > width)
                        bwidget->panx = width;

                LOG(("panning right %d", bwidget->panx));

                fb_plotters_move_block(x + bwidget->panx, y,
                                       width - bwidget->panx, height,
                                       x, y);

                bwidget->scrollx += bwidget->panx;
                fb_queue_redraw(widget, width - bwidget->panx, 0, width, height);
        }


        bwidget->pan_required = false;
        bwidget->panx = 0;
        bwidget->pany = 0;
}

static void fb_redraw(fbtk_widget_t *widget,
                      struct browser_widget_s *bwidget,
                      struct browser_window *bw)
{
        struct content *c;
        int x;
        int y;
        int width;
        int height;

        c = bw->current_content;

        if ((!c) || (c->locked))
                return;

        height = fbtk_get_height(widget);
        width = fbtk_get_width(widget);
        x = fbtk_get_x(widget);
        y = fbtk_get_y(widget);

        /* adjust clipping co-ordinates according to window location */
        bwidget->redraw_box.y0 += y;
        bwidget->redraw_box.y1 += y;
        bwidget->redraw_box.x0 += x;
        bwidget->redraw_box.x1 += x;


        /* redraw bounding box is relative to window */
        content_redraw(c,
                       x - bwidget->scrollx, y - bwidget->scrolly,
                       width, height,
                       bwidget->redraw_box.x0, bwidget->redraw_box.y0,
                       bwidget->redraw_box.x1, bwidget->redraw_box.y1,
                       bw->scale, 0xFFFFFF);


        fb_os_redraw(&bwidget->redraw_box);

        bwidget->redraw_box.y0 = bwidget->redraw_box.x0 = INT_MAX;
        bwidget->redraw_box.y1 = bwidget->redraw_box.x1 = -(INT_MAX);
        bwidget->redraw_required = false;
}

static int
fb_browser_window_redraw(fbtk_widget_t *widget, void *pw)
{
        struct gui_window *gw = pw;
        struct browser_widget_s *bwidget;

        bwidget = fbtk_get_userpw(widget);

        if (bwidget->pan_required) {
                int pos;
                fb_pan(widget, bwidget, gw->bw);
                pos = (bwidget->scrollx * 100) / gw->bw->current_content->width;;
                fbtk_set_scroll_pos(gw->hscroll, pos);
                pos = (bwidget->scrolly * 100) / gw->bw->current_content->height;
                fbtk_set_scroll_pos(gw->vscroll, pos);

        }

        /*if (bwidget->redraw_required)*/ {
                fb_redraw(widget, bwidget, gw->bw);
        }
        return 0;
}

static void *myrealloc(void *ptr, size_t len, void *pw)
{
        return realloc(ptr, len);
}

void gui_init(int argc, char** argv)
{
        char buf[PATH_MAX];

        LOG(("argc %d, argv %p", argc, argv));

        fb_find_resource(buf, "Aliases", "./framebuffer/res/Aliases");
        LOG(("Using '%s' as Aliases file", buf));
        if (hubbub_initialise(buf, myrealloc, NULL) !=
                        HUBBUB_OK)
                die("Unable to initialise HTML parsing library.\n");

        /* load browser messages */
        fb_find_resource(buf, "messages", "./framebuffer/res/messages");
        LOG(("Using '%s' as Messages file", buf));
        messages_load(buf);

        /* load browser options */
        fb_find_resource(buf, "Options", "~/.netsurf/Options");
        LOG(("Using '%s' as Preferences file", buf));
        options_file_location = strdup(buf);
        options_read(buf);
        
        /* setup some options */
        verbose_log = option_fb_log_enabled;
        
        /* load browser bookmarks */
        fb_find_resource(buf, "Bookmarks", "~/.netsurf/Bookmarks");
        LOG(("Using '%s' as Bookmarks file", buf));
        bookmarks_read(buf);

        /* set up stylesheet urls */
        fb_find_resource(buf, "default.css", "./framebuffer/res/default.css");
        default_stylesheet_url = path_to_url(buf);
        LOG(("Using '%s' as Default CSS URL", default_stylesheet_url));

        framebuffer = fb_os_init(argc, argv);
                if(!framebuffer)
                        die("Error initialising framebuffer");

        fb_os_option_override();

        option_target_blank = false;

        switch (framebuffer->bpp) {
                /*        case 1:
                plot = framebuffer_1bpp_plot;
                break;
                */
        case 8:
                plot = framebuffer_8bpp_plot;
                break;

        case 16:
                plot = framebuffer_16bpp_plot;
                break;

        case 32:
                plot = framebuffer_32bpp_plot;
                break;

        default:
                LOG(("Unsupported bit depth (%d)", framebuffer->bpp));
                die("Unsupported bit depth");
        }

        framebuffer->cursor = fb_cursor_init(framebuffer, &pointer_image);

        if (fb_font_init() == false)
                die("Unable to initialise the font system");

        fbtk = fbtk_init(framebuffer);
}

void gui_init2(int argc, char** argv)
{
        struct browser_window *bw;
        const char *addr = NETSURF_HOMEPAGE;

        LOG(("argc %d, argv %p", argc, argv));

        if (option_homepage_url != NULL && option_homepage_url[0] != '\0')
                addr = option_homepage_url;

        if (argc > 1) addr = argv[1];

        LOG(("calling browser_window_create"));
        bw = browser_window_create(addr, 0, 0, true, false);
        
        if(bw && bw->window)
                fbtk_set_focus(bw->window->browser);
        
        fb_vkb_init(bw->window);
}


void gui_multitask(void)
{
    //    LOG(("gui_multitask"));
}


void gui_poll(bool active)
{
    //    LOG(("enter fetch_poll"));
    if (active)
        fetch_poll();

    active = schedule_run() | active | redraws_pending;

    fb_os_input(fbtk, active);

    fbtk_redraw(fbtk);

}

void gui_quit(void)
{
        LOG(("gui_quit"));
        /* finish bookmarks */
        bookmarks_save();
        bookmarks_free();
        /* finish os-specific stuff (video etc.) */
        fb_os_quit(framebuffer);
        /* We don't care if this fails as we're about to die, anyway */
        hubbub_finalise(myrealloc, NULL);
}



static void
fb_update_back_forward(struct gui_window *gw)
{
        struct browser_window *bw = gw->bw;

        fbtk_set_bitmap(gw->back, 
                        (browser_window_back_available(bw)) ?
                                        &left_arrow : &left_arrow_g);
        fbtk_set_bitmap(gw->forward, 
                        (browser_window_forward_available(bw)) ?
                                        &right_arrow : &right_arrow_g);
}

static int 
fb_forwardmenu_select(fbtk_widget_t *widget, int sel, void *pw)
{
        LOG(("sel=%d", sel));
        struct gui_window *gw =pw;
        // go fwd (sel) times
        struct history* h = gw->bw->history; 
        struct history_entry* he = history_get_current(h);
        for(int i = 0; i < sel; i++) {
                he = history_get_forward(h, he);
                if(!he) return true; 
        }
        // direkte Navigation mit hilfe der History
        history_go(gw->bw, h, he, false);
        
                // the menu has overlapped the browser
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        
        // Always close the menu for now. ('false' means keep it open) 
        return true;         
}

static void
fb_popup_forward_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, int screenx, int screeny)
{
        char menutext[10000] = {0}; 
        struct history* h = gw->bw->history; 
        struct history_entry* he = history_get_current(h);
        
        he = history_get_forward(h, he);
        if(!he) return; 
        
        for(int i = 0; i < 10; i++) {
                if(!he) break; 
                strcat(menutext, history_get_title(h, he));
                strcat(menutext, "\n");
                he = history_get_forward(h, he);
        }
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_forwardmenu_select, gw);        
}

static int 
fb_backmenu_select(fbtk_widget_t *widget, int sel, void *pw)
{
        LOG(("sel=%d", sel));
        struct gui_window *gw =pw;
        // go back (sel) times
        struct history* h = gw->bw->history; 
        struct history_entry* he = history_get_current(h);
        for(int i = 0; i < sel; i++) {
                he = history_get_back(h, he);
                if(!he) return true; 
        }
        // direkte Navigation mit hilfe der History
        history_go(gw->bw, h, he, false);
        
        
        // the menu has overlapped the browser
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        
        // Always close the menu for now. ('false' means keep it open) 
        return true;         
}

static void
fb_popup_back_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, int screenx, int screeny)
{
        char menutext[10000] = {0}; 
        struct history* h = gw->bw->history; 
        struct history_entry* he = history_get_current(h);
        
        he = history_get_back(h, he);
        if(!he) return; 
        
        for(int i = 0; i < 10; i++) {
                if(!he) break; 
                strcat(menutext, history_get_title(h, he));
                strcat(menutext, "\n");
                he = history_get_back(h, he);
        }
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_backmenu_select, gw);
}

static fb_bookmark_group_t* fb_bookmark_nth_group(int n) // n counts from 0, which is the first group
{
        if(n < 0) return NULL; 
        
        fb_bookmark_group_t* g = bookmarks_get();
        for(int i = 0; i < n; i++) {
                if(!g) break;
                g = g->next;
        }
        
        return g; 
}

static fb_bookmark_t* fb_bookmark_nth(int n, fb_bookmark_group_t* g) // n counts from 0, which is the first group
{
        if(n < 0) return NULL; 
        
        fb_bookmark_t* e = g->entry;
        for(int i = 0; i < n; i++) {
                if(!e) break;
                e = e->next;
        }
        
        return e; 
}

static int
fb_goto_bk_select(fbtk_widget_t *widget, int sel, void *pw)
{
        fb_bookmark_group_t* g = pw; 
        fb_bookmark_t* e = fb_bookmark_nth(sel-1, g);
        
        LOG(("sel=%d url=%s", sel, (e ? e->url : NULL)));
        // get gui_window
        // FIXME: we need to pass this via some other pointer entry of the menu
        struct gui_window* gw = window_list;
                
        if(e)
        {
                // goto bookmarked page
                struct browser_window *bw = gw->bw;
                browser_window_go(bw, e->url, 0, true);
        }
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        // Close the menu.
        return true; 
}

static void
fb_popup_goto_bk_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, fb_bookmark_group_t* g, int screenx, int screeny)
{
        char menutext[10000] = {0}; 
        if(!g) return; 
        fb_bookmark_t* e = g->entry;
        if(!e) return; 
        
        for(; e; e = e->next)
        {
                strncat(menutext, e->title, 80);
                strcat(menutext, "\n");
        }
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_goto_bk_select, g);
}


static int
fb_goto_bkgroups_select(fbtk_widget_t *widget, int sel, void *pw)
{
        struct gui_window *gw =pw;
        fb_bookmark_group_t* g = fb_bookmark_nth_group(sel-1);
        
        LOG(("sel=%d name=%s", sel, (g ? g->name : NULL)));
        if(g) {
                bbox_t box; 
                fbtk_menu_get_bbox(widget, sel, &box);
                int screenx = box.x1 + fbtk_get_x(widget);
                int screeny = box.y0 + fbtk_get_y(widget); 
                
                // popup submenu with the bookmarks contained. 
                fb_popup_goto_bk_menu(gw, widget, g, screenx, screeny); 
                // Don't close the menu! We have opened a submenu. 
                return false; 
        }
        else {
                // redraw complete browser contents to fix graphics (for now) 
                fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
                // Close the menu.
                return true; 
        }
}

static void
fb_popup_goto_bkgroups_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, int screenx, int screeny)
{
        char menutext[10000] = {0}; 
        fb_bookmark_group_t* g = bookmarks_get();
        if(!g) return; 
        
        for(; g; g = g->next)
        {
                strcat(menutext, g->name);
                if(g->entry) strcat(menutext, "  ->");
                strcat(menutext, "\n");
        }
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_goto_bkgroups_select, gw);
}


static int
fb_set_bkgroups_select(fbtk_widget_t *widget, int sel, void *pw)
{
        struct gui_window *gw =pw;
        fb_bookmark_group_t* g = fb_bookmark_nth_group(sel-1);
        
        LOG(("sel=%d name=%s", sel, (g ? g->name : NULL)));
        if(g) {
                // get current page information
                struct history* h = gw->bw->history; 
                struct history_entry* he = history_get_current(h);
                const char *url = history_get_url(h, he);
                const char *title = history_get_title(h, he); 
                // set (add) the bookmark 
                bookmarks_add(g, url, title);
        }
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        // Close the menu.
        return true; 
}

static void
fb_popup_set_bkgroups_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, int screenx, int screeny)
{
        char menutext[10000] = {0}; 
        fb_bookmark_group_t* g = bookmarks_get();
        if(!g) return; 
        
        for(; g; g = g->next)
        {
                strcat(menutext, g->name);
                strcat(menutext, "\n");
        }
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_set_bkgroups_select, gw);
}


/*
// there are several problems with zooming
// - font size doesn't change
// - images are not scaled
// - link positions (for mouse hover etc.) are not updated 

static int
fb_set_zoom_select(fbtk_widget_t *widget, int sel, void *pw)
{
        struct gui_window *gw =pw;
        if(sel >= 0) {
                const char* tx = fbtk_menu_get_text(widget, sel);
                if(!tx)
                        tx = "100";
                int percent = atoi(tx);
                percent = max(percent, 10);
                percent = min(percent, 1000);
                
                LOG(("sel=%d zoom to %d%% tx=%s", sel, percent, tx));
                browser_window_set_scale(gw->bw, percent * 0.01, true);
        }
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        // Close the menu.
        return true; 
}

static void 
fb_popup_zoom_menu(struct gui_window* gw, fbtk_widget_t* menu_parent, int screenx, int screeny)
{
        const char* menutext = "50 %\n70 %\n80 %\n90 %\n100 %\n110 %\n120 %\n150 %\n180 %\n200 %"; 
        fbtk_popup_menu(gw->window, menu_parent, screenx, screeny, FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                        menutext, fb_set_zoom_select, gw);
}
*/

static int
fb_mainmenu_select(fbtk_widget_t *widget, int sel, void *pw)
{
        struct gui_window *gw =pw;
        LOG(("sel=%d", sel));
        
        bbox_t box; 
        fbtk_menu_get_bbox(widget, sel, &box);
        int screenx = box.x1 + fbtk_get_x(widget);
        int screeny = box.y0 + fbtk_get_y(widget); 
        
        // action!
        switch(sel) {
                case 1: // Go to..
                        fbtk_set_focus(gw->url);
                        break;
                case 2: // Search..
                {
                        const char* search_url = option_fb_search_url; 
                        if(!search_url || !strlen(search_url))
                                search_url = FB_DEFAULT_SEARCH_URL;
                        fbtk_set_text(gw->url, search_url);
                        fbtk_set_focus(gw->url);
                        break; 
                }
                case 3: // back
                        /*if (history_back_available(gw->bw->history))
                                history_back(gw->bw, gw->bw->history);
                        fb_update_back_forward(gw); */
                        fb_popup_back_menu(gw, widget, screenx, screeny);
                        return false; // keep this menu open
                
                case 4: // fwd
                        /*if (history_forward_available(gw->bw->history))
                                history_forward(gw->bw, gw->bw->history);
                        fb_update_back_forward(gw); */
                        fb_popup_forward_menu(gw, widget, screenx, screeny); 
                        return false; // keep this menu open
                                
                case 5: // home page
                {
                        const char *addr = NETSURF_HOMEPAGE;
                        struct browser_window *bw = gw->bw;

                        if (option_homepage_url != NULL && option_homepage_url[0] != '\0')
                                addr = option_homepage_url;

                        browser_window_go(bw, addr, 0, true);
                        break;
                }   
                                     
                case 6: // goto bookmark
                        fb_popup_goto_bkgroups_menu(gw, widget, screenx, screeny);
                        return false; 
                case 7: // set bookmark in... 
                        fb_popup_set_bkgroups_menu(gw, widget, screenx, screeny); 
                        return false; 
                        
                /*case 8: // zoom
                        fb_popup_zoom_menu(gw, widget, screenx, screeny);
                        return false; */
                        
                case 8: // help
                {
                        const char *addr = FB_HELP_URL;
                        struct browser_window *bw = gw->bw;

                        if (option_fb_help_url != NULL && option_fb_help_url[0] != '\0')
                                addr = option_fb_help_url;

                        browser_window_go(bw, addr, 0, true);
                        break;
                }
                
                default:
                        break;
        }
        
        // the menu has overlapped the browser
        // redraw complete browser contents to fix graphics (for now) 
        fb_queue_redraw(gw->browser, 0, 0, fbtk_get_width(gw->browser), fbtk_get_height(gw->browser));
        // Always close the menu for now. ('false' means keep it open) 
        return true; 
}
static void
fb_popup_mainmenu(struct gui_window* gw)
{
                fbtk_popup_menu(gw->window, gw->browser, 0, fbtk_get_y(gw->browser), FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                                "Go to...\nSearch...\nBack     ->\nForward ->\nHome page\n"
                                "Goto Bookmark ->\nSet Bookmark in ->\nHelp", fb_mainmenu_select, gw);
}                

/* called back when click in browser window */
static int
fb_browser_window_click(fbtk_widget_t *widget,
                        browser_mouse_state st,
                        int x, int y,
                        void *pw)
{
        struct browser_window *bw = pw;
        struct browser_widget_s *bwidget = fbtk_get_userpw(widget);

        LOG(("browser window clicked at %d,%d st=%X",x,y,st));
        browser_window_mouse_click(bw,
                                   st,
                                   x + bwidget->scrollx,
                                   y + bwidget->scrolly);
        return 0;
}

/* called back when movement in browser window */
static int
fb_browser_window_move(fbtk_widget_t *widget,
                        int x, int y,
                        void *pw)
{
        struct browser_window *bw = pw;
        struct browser_widget_s *bwidget = fbtk_get_userpw(widget);

        browser_window_mouse_track(bw,
                                   0,
                                   x + bwidget->scrollx,
                                   y + bwidget->scrolly);

        return 0;
}


#define FB_LINK_SECONDARY_MULT 2
struct fb_link_nav_data
{
        int search_x;
        int search_y;
        enum fb_nav_dir search_dir; 
        
        struct box* best_box;
        int best_dist; 
        int best_x;
        int best_y; 
};

/** Computes the distance of the range c1..c2 from ref.
  * If c1 <= ref <= c2, returns 0. 
  * Returns a negative number if ref &lt; c1. 
  */
static int
fb_dist(int ref, int c1, int c2)
{
        if(ref <= c1)
                return ref - c1 - 1; 
        if(ref >= c2)
                return ref - c2 + 1; 
        // it's in between. 
        return 0; 
}

static void
fb_browser_box_enum(struct box* b, int boxx, int boxy, void* pd)
{
        assert(pd);
        struct fb_link_nav_data* data = pd; 
        
        if(b && (b->href || b->gadget)) {
                
                int xdist = fb_dist(data->search_x, boxx, boxx + b->width);
                int ydist = fb_dist(data->search_y, boxy, boxy + b->height);
                
                
                int prim_dist, sec_dist; 
                if(data->search_dir == FB_NAV_UP || data->search_dir == FB_NAV_DOWN) {
                        prim_dist = ydist; sec_dist = abs(xdist); 
                }
                else {
                        prim_dist = xdist; sec_dist = abs(ydist); 
                }
                if(data->search_dir == FB_NAV_DOWN || data->search_dir == FB_NAV_RIGHT) {
                        // need higher things, but fb_dist returns negative numbers for them.
                        prim_dist = -prim_dist;  
                }
                
                if(prim_dist <= 0 || sec_dist > prim_dist) {
                        //LOG(("wrong direction"));
                        return;
                }
                if(b->width <= 0 || b->height <= 0) {
                        // link has zero size
                        // you can't hit it using the mouse
                        // so we don't have to check it either
                        return; 
                }
                
                /*LOG(("href found at %d,%d size=(%d,%d) target %s", boxx, boxy, 
                         b->width, b->height, b->href));
                LOG(("xdist=%d ydist=%d", xdist, ydist));
                LOG(("prim_dist=%d sec_dist=%d", prim_dist, sec_dist));*/
                
                // the elmenent is in the right direction
                // compute its distance and use it if better. 
                int dist = prim_dist + sec_dist * FB_LINK_SECONDARY_MULT;
                if(dist < data->best_dist) {
                        data->best_box = b;
                        data->best_dist = dist; 
                        data->best_x = boxx;
                        data->best_y = boxy; 
                }
        }
}

static bool
fb_window_link_nav(struct gui_window* gw, enum fb_nav_dir dir)
{
        int pointer_x, pointer_y; 
        fb_os_get_pointer_pos(&pointer_x, &pointer_y);
        int scroll_x, scroll_y;
        gui_window_get_scroll(gw, &scroll_x, &scroll_y);
        
        struct fb_link_nav_data data;
        data.search_x = pointer_x + scroll_x - fbtk_get_x(gw->browser);
        data.search_y = pointer_y + scroll_y - fbtk_get_y(gw->browser);
        data.search_dir = dir; 
        data.best_dist = INT_MAX;
        data.best_box = NULL; 
        LOG(("searching from %d,%d; dir %d", data.search_x, data.search_y, dir));
        
        browser_window_enum_boxes(gw->bw, &fb_browser_box_enum, &data);
        
        if(data.best_box) {
                LOG(("best_box found, links to %s", data.best_box->href));
                int posx = data.best_x + data.best_box->width/2, 
                          posy = data.best_y + 5;
                gui_window_scroll_visible(gw, data.best_x, data.best_y, 
                                                                  data.best_x + data.best_box->width,
                                                                    data.best_y + data.best_box->height);
                LOG(("pointing to %d,%d", posx, posy));
        
                gui_window_get_scroll_with_pan(gw, &scroll_x, &scroll_y);
                fb_os_warp_pointer(posx - scroll_x + fbtk_get_x(gw->browser), 
                                                   posy - scroll_y + fbtk_get_y(gw->browser));
        }
        
        return (data.best_box != NULL);         
}

static void
fb_update_navmode(struct gui_window *gw)
{
        struct bitmap* bm = NULL;
        switch(fb_nav_mode) {
                default: 
                case FB_NAV_LINK: 
                        bm = &nav_link_image; break;
                case FB_NAV_SCROLL: 
                        bm = &nav_scroll_image; break;
                case FB_NAV_SCROLL_PAGE:
                        bm = &nav_scroll_page_image; break; 
                case FB_NAV_MOVE_POINTER: 
                        bm = &nav_move_pointer_image; break;
        }
        fbtk_set_bitmap(gw->navmode, bm);
        
}

static void
fb_toggle_navmode(struct gui_window *gw)
{
        // switch nav mode
        switch(fb_nav_mode) {
                case FB_NAV_LINK:
                        fb_nav_mode = FB_NAV_SCROLL_PAGE;
                        break;
                case FB_NAV_SCROLL_PAGE:
                        fb_nav_mode = FB_NAV_SCROLL;
                        break;
                case FB_NAV_SCROLL:
                        fb_nav_mode = FB_NAV_MOVE_POINTER;
                        break;
                default:
                case FB_NAV_MOVE_POINTER:
                        fb_nav_mode = FB_NAV_LINK;
                        break;
        }
        fb_update_navmode(gw);
}

static void
fb_move_pointer(int dx, int dy)
{
        int x,y;
        fb_os_get_pointer_pos(&x, &y);
        x += dx; y += dy; 
        
        x = max(x, 0); y = max(y, 0);
        x = min(x, framebuffer->width);
        y = min(y, framebuffer->height); 
        
        fb_os_warp_pointer(x, y);
}

static fbtk_input_t fb_old_url_handler_input = NULL; 
static int
fb_url_window_input(fbtk_widget_t *widget, int value, void *pw)
{
        struct gui_window *gw = pw;
        int res = 0; 
        
        if(gw->vkb && value == 'z')
                value = '\r'; 
        
        switch(value) {
        case FB_KEY_MENU:
                fb_popup_mainmenu(gw);
                break;
        case FB_KEY_NAVMODE: 
                if(gw->vkb)
                        fb_toggle_vkb_mode(gw);
                else
                        fb_toggle_navmode(gw);
                break;
                
        default: 
                res = fb_old_url_handler_input(widget, value, pw);
                break;
        }
        return res; 
}

static int
fb_browser_window_input(fbtk_widget_t *widget, int value, void *pw)
{
        struct gui_window *gw = pw;
        int res = 0;
        int scrolly, scrollx; 
        if(fb_nav_mode == FB_NAV_SCROLL_PAGE) {
                scrolly = fbtk_get_height(gw->browser) - 50; 
                scrollx = fbtk_get_width(gw->browser) - 50; 
        }
        else {
                scrolly = 30;
                scrollx = 30; 
        }
        
        if(gw->vkb && value == 'z')
                value = '\r'; 
        
        LOG(("got value %d",value));
        switch (value) {

        case KEY_PAGE_UP:
                fb_window_scroll(gw, 0, -fbtk_get_height(gw->browser));
                break;

        case KEY_PAGE_DOWN:
                fb_window_scroll(gw, 0, fbtk_get_height(gw->browser));
                break;

        case KEY_RIGHT:
		if(fb_nav_mode == FB_NAV_LINK) {
			if(fb_window_link_nav(gw, FB_NAV_RIGHT))
				break;
		}
		if(fb_nav_mode == FB_NAV_SCROLL || fb_nav_mode == FB_NAV_SCROLL_PAGE)
                	fb_window_scroll(gw, scrollx, 0);
		else
                        fb_move_pointer(FB_MOVE_POINTER_DIST, 0);

                break;

        case KEY_LEFT:
		if(fb_nav_mode == FB_NAV_LINK) {
			if(fb_window_link_nav(gw, FB_NAV_LEFT))
				break;
		}
                if(fb_nav_mode == FB_NAV_SCROLL || fb_nav_mode == FB_NAV_SCROLL_PAGE)
                        fb_window_scroll(gw, -scrollx, 0);
                else
                        fb_move_pointer(-FB_MOVE_POINTER_DIST, 0);
                break;

        case KEY_UP:
		if(fb_nav_mode == FB_NAV_LINK) {
			if(fb_window_link_nav(gw, FB_NAV_UP))
				break;
		}
                if(fb_nav_mode == FB_NAV_SCROLL || fb_nav_mode == FB_NAV_SCROLL_PAGE)
                        fb_window_scroll(gw, 0, -scrolly);
                else
                        fb_move_pointer(0, -FB_MOVE_POINTER_DIST);
                break;

        case KEY_DOWN:
		if(fb_nav_mode == FB_NAV_LINK) {
			if(fb_window_link_nav(gw, FB_NAV_DOWN))
				break;
		}
                if(fb_nav_mode == FB_NAV_SCROLL || fb_nav_mode == FB_NAV_SCROLL_PAGE)
                        fb_window_scroll(gw, 0, scrolly);
                else
                        fb_move_pointer(0, FB_MOVE_POINTER_DIST);
                break;
                
        case FB_KEY_MENU:
                fb_popup_mainmenu(gw);
                break;
                
	case FB_KEY_HOME:
	{
		const char *addr = NETSURF_HOMEPAGE;
		struct browser_window *bw = gw->bw;

		if (option_homepage_url != NULL && option_homepage_url[0] != '\0')
				addr = option_homepage_url;

		browser_window_go(bw, addr, 0, true);
		break;
	}
                                        
	case FB_KEY_BACK:
		if (history_back_available(gw->bw->history))
			history_back(gw->bw, gw->bw->history);                        
		fb_update_back_forward(gw);
		break;
                
        case FB_KEY_NAVMODE:
                if(fb_is_writing_text(gw) && gw->vkb)
                        fb_toggle_vkb_mode(gw);
                else
                        fb_toggle_navmode(gw);
                break;
                
        case '\r':
        {
                int px, py; 
                fb_os_get_pointer_pos(&px, &py);
                fbtk_widget_t* root = fbtk_get_root_widget(widget);
                LOG(("RETURN @pos: %d,%d", px, py));
                fbtk_click(root, BROWSER_MOUSE_PRESS_1); // press
                fbtk_click(root, BROWSER_MOUSE_CLICK_1); // and release!
                break;
        }

        default:
                res = browser_window_key_press(gw->bw, value);
                break;
        }

        return 0;
}
/* left icon click routine */
static int
fb_leftarrow_click(fbtk_widget_t *widget,
                   browser_mouse_state st,
                   int x, int y, void *pw)
{
        struct gui_window *gw =pw;
        struct browser_window *bw = gw->bw;

        if (st == BROWSER_MOUSE_CLICK_1) {
                if (history_back_available(bw->history))
                        history_back(bw, bw->history);
        }
        fb_update_back_forward(gw);
        return 0;

}

/* right arrow icon click routine */
static int
fb_rightarrow_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        struct gui_window *gw =pw;
        struct browser_window *bw = gw->bw;

        if (st == BROWSER_MOUSE_CLICK_1) {
                if (history_forward_available(bw->history))
                        history_forward(bw, bw->history);
        }
        fb_update_back_forward(gw);
        return 0;

}

/* reload icon click routine */
static int
fb_reload_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        struct browser_window *bw = pw;
        browser_window_reload(bw, true);
        return 0;
}

/* stop icon click routine */
static int
fb_stop_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        struct browser_window *bw = pw;
        browser_window_stop(bw);
        return 0;
}

/* left scroll icon click routine */
static int
fb_scrolll_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        fbtk_input(widget, KEY_LEFT);
        return 0;
}

static int
fb_scrollr_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        fbtk_input(widget, KEY_RIGHT);
        return 0;
}

static int
fb_scrollu_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        fbtk_input(widget, KEY_UP);
        return 0;
}

static int
fb_scrolld_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        fbtk_input(widget, KEY_DOWN);
        return 0;
}

static int
fb_url_enter(void *pw, char *text)
{
        struct browser_window *bw = pw;
        browser_window_go(bw, text, 0, true);
        fbtk_set_focus(bw->window->browser);
        return 0;
}

static int
fb_url_move(fbtk_widget_t *widget,
            int x, int y,
            void *pw)
{
        fb_cursor_set(framebuffer->cursor, &caret_image);
        return 0;
}

static int
set_ptr_default_move(fbtk_widget_t *widget,
            int x, int y,
            void *pw)
{
        fb_cursor_set(framebuffer->cursor, &pointer_image);
        return 0;
}

static int
set_ptr_hand_move(fbtk_widget_t *widget,
            int x, int y,
            void *pw)
{
        fb_cursor_set(framebuffer->cursor, &hand_image);
        return 0;
}


static int
fb_menu_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        // act only when LMB clicked
        if(!(st & BROWSER_MOUSE_CLICK_1)) return 0; 
        
        struct gui_window *gw =pw;
        fb_popup_mainmenu(gw);
        return 0;
}

static int
fb_navmode_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        // act only when LMB clicked
        if(!(st & BROWSER_MOUSE_CLICK_1)) return 0; 
        
        struct gui_window *gw =pw;
        fb_toggle_navmode(gw);
        
        return 0;
}

// --------------------------------------------------------------------------------------------------
// VKB begin

static int fb_vkb_default_output[FB_VKB_NUM_KEYS][FB_VKB_NUM_ITEMS] = 
{
        { ' ', '.', ',', '!', '?' },    // 0
        { ':', '/', '@', '$', '&' },    // 1
        { 'a', 'b', 'c'},               // 2
        { 'd', 'e', 'f'},               // 3
        { 'g', 'h', 'i'},
        { 'j', 'k', 'l'},
        { 'm', 'n', 'o'},
        { 'p', 'q', 'r', 's'},
        { 't', 'u', 'v'},
        { 'w', 'x', 'y', 'z'}
}; 

static void fb_vkb_parse(int* pout, const char* pstr)
{
        if(!pstr || !strlen(pstr)) return; 
        
        memset(pout, 0, sizeof(int)*FB_VKB_NUM_ITEMS);
        
        uint32_t ucs4;
        size_t nxtchr = 0;
        int outidx = 0; 
        size_t length = strlen(pstr);
        
        while (nxtchr < length) {
                ucs4 = utf8_to_ucs4(pstr + nxtchr, length - nxtchr);
                nxtchr = utf8_next(pstr, length, nxtchr);
                pout[outidx] = ucs4; 
                outidx++;
                if(outidx >= FB_VKB_NUM_ITEMS) break; 
        }
        LOG(("%c,%c,%c,%c", (char)pout[0], (char)pout[1], (char)pout[2], (char)pout[3])); 
}

void fb_vkb_init(struct gui_window* gw)
{
        vkb_state_t* s = &gw->vkb_state;
        s->mode = FB_VKB_123;
        s->current_key = -1; 
        s->current_item = -1; 
        memcpy(s->output, fb_vkb_default_output, sizeof(fb_vkb_default_output));
        s->timeout_ms = max(option_fb_vkb_timeout_ms, 100); 
        for(int i = 0; i < FB_VKB_NUM_KEYS; i++)
                fb_vkb_parse(s->output[i], option_fb_vkb[i]);
}

int fb_vkb_get_current(struct gui_window* gw)
{
        vkb_state_t* s = &gw->vkb_state;
        if(s->current_key < 0 || s->current_key >= FB_VKB_NUM_KEYS) return 0;
        if(s->current_item < 0 || s->current_item >= FB_VKB_NUM_ITEMS) return 0; 
        int r = s->output[s->current_key][s->current_item]; 
        if(s->mode == FB_VKB_ABC || s->mode == FB_VKB_Abc)
                r = towupper(r);
        else if(s->mode == FB_VKB_abc)
                r = towlower(r);
        LOG(("curKey=%d curItem=%d r=%c (0x%X)", s->current_key, s->current_item, (char)r, r));
        return r; 
}

int fb_vkb_input(struct gui_window* gw, int ch)
{
        vkb_state_t* s = &gw->vkb_state;
        
        LOG(("ch=%c (0x%X)", (char)ch, ch));
        if(ch < '0' || ch > '9') {
                // another character, using a not-vkb-mapped key. we don't care for it. 
                // but reset vkb state. 
                // FIXME: the return value should not be ignored
                // but instead be inserted additionally, before ch. 
                fb_vkb_timeout(gw); 
                // return char unchanged
                return ch; 
        }
        
        // it's one of our keys!
        // map ascii value to VKB key
        int key = ch - '0'; 
        // now what will we do... 
        if(s->mode == FB_VKB_123) {
                fb_vkb_timeout(gw); 
                return ch; // numbers: return unchanged
        }
        
        int ret = 0; 
        
        if(key != s->current_key) {
                // input last key, if any
                if(s->current_key >= 0)
                        ret = fb_vkb_timeout(gw);
                // setup new key
                s->current_key = key;
                s->current_item = 0; 
        }
        else {
                s->current_item++;
                if(s->current_item >= FB_VKB_NUM_ITEMS || !s->output[s->current_key][s->current_item])
                        s->current_item = 0; 
        }
        
        fbtk_request_redraw(gw->vkb);
        return ret; 
}

int fb_vkb_timeout(struct gui_window* gw)
{
        vkb_state_t* s = &gw->vkb_state;
        // remember current character
        int ret = fb_vkb_get_current(gw);
        
        // may need to change mode
        if(s->mode == FB_VKB_Abc) s->mode = FB_VKB_abc; 
        
        // reset current character
        s->current_key = -1; 
        s->current_item = -1; 
        fbtk_request_redraw(gw->vkb);
        
        LOG(("ret=%c (0x%X)", (char)ret, ret));
        // return cur char
        return ret; 
}

void fb_toggle_vkb_mode(struct gui_window* gw)
{
        gw->vkb_state.mode = (gw->vkb_state.mode + 1) % FB_VKB_MODE_COUNT;
        fbtk_request_redraw(gw->vkb);
}

static int fb_vkb_click(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw)
{
        if(!(st & BROWSER_MOUSE_CLICK_1)) return 0; 
        
        struct gui_window *gw =pw;
        fb_toggle_vkb_mode(gw);
        
        return 0;
}


static int fb_vkb_redraw(fbtk_widget_t *widget, void *pw)
{
        struct gui_window* gw = pw; 
        
        /* clear background */
        /* transparent polygon filling isnt working so fake it */
        plot.fill(fb_plot_ctx.x0, fb_plot_ctx.y0,
                        fb_plot_ctx.x1, fb_plot_ctx.y1,
                        FB_FRAME_COLOUR);

        plot.rectangle(fb_plot_ctx.x0, fb_plot_ctx.y0,
                        fb_plot_ctx.x1 - fb_plot_ctx.x0 - 1,
                        fb_plot_ctx.y1 - fb_plot_ctx.y0 - 1,
                        1, 0x00000000, false, false);
        
        struct css_style text_style;
        memset(&text_style, 0, sizeof(text_style));
        text_style.font_size.value.length.unit = CSS_UNIT_PX;
        text_style.font_size.value.length.value = 12;
        text_style.font_weight = CSS_FONT_WEIGHT_BOLD;
        
        const char* mode_text = "";
        switch(gw->vkb_state.mode) {
                case FB_VKB_123: mode_text = "123"; break;
                case FB_VKB_Abc: mode_text = "Abc"; break;
                case FB_VKB_ABC: mode_text = "ABC"; break;
                case FB_VKB_abc: mode_text = "abc"; break;
        }
                
        plot.text(fb_plot_ctx.x0 + 3,
                  fb_plot_ctx.y0 + 11,
                  &text_style,
                  mode_text, strlen(mode_text),
                  FB_FRAME_COLOUR, 0xFF444444);
        
        // convert current character to UTF8 
        char char_text[10] = {0}; 
        int char_ucs = fb_vkb_get_current(gw);
        int nb = utf8_from_ucs4(char_ucs, char_text); 
        char_text[nb] = '\0'; // zero termination 
        
        text_style.font_size.value.length.value = 16;
        plot.text(fb_plot_ctx.x0 + 8,
                  fb_plot_ctx.y0 + 24,
                  &text_style,
                  char_text, strlen(char_text),
                  0x00000000, 0xFF000000);

        fb_os_redraw(&fb_plot_ctx);

        return 0;
}

bool fb_is_writing_text(struct gui_window* gw)
{
        if(!gw) return false;
        
        fbtk_widget_t* focus = fbtk_get_focused_widget(gw->window);
        if(!focus) return false; 
        
        // edit widget 
        if(fbtk_is_editable(focus)) return true; 
        
        // browser is selected and is showing a cursor
        if(focus == gw->browser && gw->c_h > 0) return true; 
        
        // 
        return false; 
}
                
// VKB end
// --------------------------------------------------------------------------------------------------


struct gui_window *
gui_create_browser_window(struct browser_window *bw,
                          struct browser_window *clone,
                          bool new_tab)
{
        struct gui_window *gw;
        struct browser_widget_s *browser_widget;
        fbtk_widget_t *widget;
        int top = 0;
        int bot = 0;
        int right = 0;
        
        int vkb_width = (option_fb_vkb_enabled ? 30 : 0);

        gw = calloc(1, sizeof(struct gui_window));

        if (gw == NULL)
                return NULL;

        /* seems we need to associate the gui window with the underlying
         * browser window
         */
        gw->bw = bw;


        switch(bw->browser_window_type) {
        case BROWSER_WINDOW_NORMAL:
                gw->window = fbtk_create_window(fbtk, 0, 0, 0, 0);

                top = 30;
                bot = 20;
                right = 18;
                LOG(("Normal window"));

                /* fill toolbar background */
                widget = fbtk_create_fill(gw->window, 
                                          0, 0, 0, 30, 
                                          FB_FRAME_COLOUR);
                fbtk_set_handler_move(widget, set_ptr_default_move, bw);

                /* added: menu button */
                gw->menu = fbtk_create_button(gw->window, 
                                            5, 2, 
                                            FB_FRAME_COLOUR, 
                                            &menu_image, 
                                            fb_menu_click, 
                                            gw);
                fbtk_set_handler_move(gw->menu, set_ptr_hand_move, bw);

                /* back button */
                gw->back = fbtk_create_button(gw->window, 
                                            35, 2, 
                                            FB_FRAME_COLOUR, 
                                            &left_arrow_g, 
                                            fb_leftarrow_click, 
                                            gw);
                fbtk_set_handler_move(gw->back, set_ptr_hand_move, bw);

                /* forward button */
                gw->forward = fbtk_create_button(gw->window, 
                                            65, 2, 
                                            FB_FRAME_COLOUR, 
                                            &right_arrow_g, 
                                            fb_rightarrow_click, 
                                            gw);
                fbtk_set_handler_move(gw->forward, set_ptr_hand_move, bw);

                /* stop button */
                widget = fbtk_create_button(gw->window, 
                                            95, 2, 
                                            FB_FRAME_COLOUR, 
                                            &stop_image, 
                                            fb_stop_click, 
                                            bw);
                fbtk_set_handler_move(widget, set_ptr_hand_move, bw);

                /* reload button */
                widget = fbtk_create_button(gw->window, 
                                            125, 2, 
                                            FB_FRAME_COLOUR, 
                                            &reload, 
                                            fb_reload_click, 
                                            bw);
                fbtk_set_handler_move(widget, set_ptr_hand_move, bw);

                /* url widget */
                gw->url = fbtk_create_writable_text(gw->window,
                                                    155 , 3,
                                                    fbtk_get_width(gw->window) - (220+vkb_width), 24,
                                                    FB_COLOUR_WHITE,
                                                    FB_COLOUR_BLACK,
                                                    true,
                                                    fb_url_enter,
                                                    bw);
                fbtk_set_handler_move(gw->url, fb_url_move, bw);
                fb_old_url_handler_input = fbtk_set_handler_input(gw->url, fb_url_window_input, gw);

                /* addition: vkb button */
                if(vkb_width) {
                        gw->vkb = fbtk_create_user(gw->window, 160 + fbtk_get_width(gw->url), 1,
                                                29, 28, gw);
                        fbtk_set_handler_move(gw->vkb, set_ptr_hand_move, bw);
                        fbtk_set_handler_redraw(gw->vkb, fb_vkb_redraw, gw);
                        fbtk_set_handler_click(gw->vkb, fb_vkb_click, gw);
                }

                /* addition: navmode button */
                gw->navmode = fbtk_create_button(gw->window, 
                                            160+vkb_width + fbtk_get_width(gw->url), 2, 
                                            FB_FRAME_COLOUR, 
                                            &nav_link_image, 
                                            fb_navmode_click, 
                                            gw);
                fbtk_set_handler_move(gw->navmode, set_ptr_hand_move, bw);

                                /* throbber */
                gw->throbber = fbtk_create_bitmap(gw->window,
                                                  190+vkb_width + fbtk_get_width(gw->url),
                                                  3,
                                                  FB_FRAME_COLOUR,
                                                  &throbber0);



                /* add status area widget, width of framebuffer less some for
                 * scrollbar
                 */
                gw->status = fbtk_create_text(gw->window,
                                              0 , 
                                              fbtk_get_height(gw->window) - bot,
                                              fbtk_get_width(gw->window) - 200 - right, 
                                              bot,
                                              FB_FRAME_COLOUR, FB_COLOUR_BLACK,
                                              false);

                fbtk_set_handler_move(gw->status, set_ptr_default_move, bw);

                /* horizontal scrollbar */
                fbtk_create_button(gw->window, 
                                   fbtk_get_width(gw->window) - 200 - right, 
                                   fbtk_get_height(gw->window) - bot, 
                                   FB_FRAME_COLOUR, 
                                   &scrolll, 
                                   fb_scrolll_click, 
                                   bw);

                fbtk_create_button(gw->window, 
                                   fbtk_get_width(gw->window) - 20 - right, 
                                   fbtk_get_height(gw->window) - bot, 
                                   FB_FRAME_COLOUR, 
                                   &scrollr, 
                                   fb_scrollr_click, 
                                   bw);

                gw->hscroll = fbtk_create_hscroll(gw->window,
                                                  fbtk_get_width(gw->window) - 160 - 20 - right,
                                                  fbtk_get_height(gw->window) - bot,
                                                  160,
                                                  bot,
                                                  FB_SCROLL_COLOUR,
                                                  FB_FRAME_COLOUR);
                /* create vertical */
                fbtk_create_button(gw->window, 
                                   fbtk_get_width(gw->window) - right, 
                                   top, 
                                   FB_FRAME_COLOUR, 
                                   &scrollu, 
                                   fb_scrollu_click, 
                                   bw);

                fbtk_create_button(gw->window, 
                                   fbtk_get_width(gw->window) - right, 
                                   fbtk_get_height(gw->window) - bot - 20, 
                                   FB_FRAME_COLOUR, 
                                   &scrolld, 
                                   fb_scrolld_click, 
                                   bw);
                
                gw->vscroll = fbtk_create_vscroll(gw->window,
                                                  fbtk_get_width(gw->window) - right,
                                                  top + 20,
                                                  right,
                                                  fbtk_get_height(gw->window) - top - bot - 40 ,
                                                  FB_SCROLL_COLOUR,
                                                  FB_FRAME_COLOUR);
                
                window_list = gw; 
                break;

        case BROWSER_WINDOW_FRAME:
                gw->window = fbtk_create_window(bw->parent->window->window, 0, 0, 0, 0);
                LOG(("create frame"));
                break;

        default:
                gw->window = fbtk_create_window(bw->parent->window->window, 0, 0, 0, 0);
                LOG(("unhandled type"));

        }

        browser_widget = calloc(1, sizeof(struct browser_widget_s));

        gw->browser = fbtk_create_user(gw->window, 0, top, -right, - (bot + top), browser_widget);

        fbtk_set_handler_click(gw->browser, fb_browser_window_click, bw);
        fbtk_set_handler_input(gw->browser, fb_browser_window_input, gw);
        fbtk_set_handler_redraw(gw->browser, fb_browser_window_redraw, gw);
        fbtk_set_handler_move(gw->browser, fb_browser_window_move, bw);

        return gw;
}

void gui_window_destroy(struct gui_window *gw)
{
        fbtk_destroy_widget(gw->window);

        free(gw);


}

void gui_window_set_title(struct gui_window *g, const char *title)
{
        LOG(("%p, %s", g, title));
}



void fb_window_scroll(struct gui_window *g, int x, int y)
{
        struct browser_widget_s *bwidget = fbtk_get_userpw(g->browser);

        bwidget->panx += x;
        bwidget->pany += y;
        bwidget->pan_required = true;

        fbtk_request_redraw(g->browser);
}

void gui_window_redraw(struct gui_window *g, int x0, int y0, int x1, int y1)
{
        fb_queue_redraw(g->browser, x0, y0, x1, y1);
}

void gui_window_redraw_window(struct gui_window *g)
{
        fb_queue_redraw(g->browser, 0, 0, fbtk_get_width(g->browser),fbtk_get_height(g->browser) );
}

void gui_window_update_box(struct gui_window *g,
                const union content_msg_data *data)
{
        fb_queue_redraw(g->browser,
                        data->redraw.x,
                        data->redraw.y,
                        data->redraw.x + data->redraw.width,
                        data->redraw.y + data->redraw.height);
}

bool gui_window_get_scroll(struct gui_window *g, int *sx, int *sy)
{
        struct browser_widget_s *bwidget = fbtk_get_userpw(g->browser);

        *sx = bwidget->scrollx;
        *sy = bwidget->scrolly;

        return true;
}

bool gui_window_get_scroll_with_pan(struct gui_window *g, int *sx, int *sy)
{
        struct browser_widget_s *bwidget = fbtk_get_userpw(g->browser);

        *sx = bwidget->scrollx + bwidget->panx;
        *sy = bwidget->scrolly + bwidget->pany;

        return true;
}

void gui_window_set_scroll(struct gui_window *g, int sx, int sy)
{
        struct browser_widget_s *bwidget = fbtk_get_userpw(g->browser);
        LOG(("scroll to %d,%d",sx,sy));
        bwidget->panx = sx - bwidget->scrollx;
        bwidget->pany = sy - bwidget->scrolly;
        bwidget->pan_required = true;

        fbtk_request_redraw(g->browser);
}

void gui_window_scroll_visible(struct gui_window *g, int x0, int y0,
                int x1, int y1)
{
        int sx, sy;
        gui_window_get_scroll(g, &sx, &sy);
        
        int bwwidth = fbtk_get_width(g->browser), 
                bwheight = fbtk_get_height(g->browser);
        
        int newsx = sx, newsy = sy; 
        newsx = max(newsx, x1-bwwidth);
        newsy = max(newsy, y1-bwheight);
        newsx = min(newsx, x0);
        newsy = min(newsy, y0);
        
        LOG(("desired: %d,%d  %d,%d", x0, y0, x1, y1));
        LOG(("current: %d,%d",sx,sy));
        
        if(newsx != sx || newsy != sy) {
                LOG(("new: %d,%d", newsx,newsy));
                gui_window_set_scroll(g, newsx, newsy); 
        }
}

void gui_window_position_frame(struct gui_window *g, int x0, int y0,
                int x1, int y1)
{
        struct gui_window *parent;
        int px, py;
        int w, h;
        LOG(("%s: %d, %d, %d, %d", g->bw->name, x0, y0, x1, y1));
        parent = g->bw->parent->window;

        if (parent->window == NULL)
                return; /* doesnt have an fbtk widget */

        px = fbtk_get_x(parent->browser) + x0;
        py = fbtk_get_y(parent->browser) + y0;
        w = x1 - x0;
        h = y1 - y0;
        if (w > (fbtk_get_width(parent->browser) - px))
            w = fbtk_get_width(parent->browser) - px;

        if (h > (fbtk_get_height(parent->browser) - py))
            h = fbtk_get_height(parent->browser) - py;

        fbtk_set_pos_and_size(g->window, px, py , w , h);

        fbtk_request_redraw(parent->browser);

}

void gui_window_get_dimensions(struct gui_window *g, int *width, int *height,
                bool scaled)
{
        *width = fbtk_get_width(g->browser);
        *height = fbtk_get_height(g->browser);
}

void gui_window_update_extent(struct gui_window *g)
{
        int pcth = 0, pctv = 0;
        
        if(g->browser && g->bw && g->bw->current_content) {
                int w = g->bw->current_content->width, 
                    h = g->bw->current_content->height;
                
                if(w>0) pcth = (fbtk_get_width(g->browser) * 100) / w;
                if(h>0) pctv = (fbtk_get_height(g->browser) * 100) / h;
        }
        fbtk_set_scroll(g->hscroll, pcth);
        fbtk_set_scroll(g->vscroll, pctv);

}

void gui_window_set_status(struct gui_window *g, const char *text)
{
        fbtk_set_text(g->status, text);
}

void gui_window_set_pointer(struct gui_window *g, gui_pointer_shape shape)
{
        switch (shape) {
        case GUI_POINTER_POINT:
                fb_cursor_set(framebuffer->cursor, &hand_image);
                break;

        case GUI_POINTER_CARET:
                fb_cursor_set(framebuffer->cursor, &caret_image);
                break;

        default:
                fb_cursor_set(framebuffer->cursor, &pointer_image);
        }
}

void gui_window_hide_pointer(struct gui_window *g)
{
}

void gui_window_set_url(struct gui_window *g, const char *url)
{
        fbtk_set_text(g->url, url);
}

static void
throbber_advance(void *pw)
{
        struct gui_window *g = pw;
        struct bitmap *image;

        switch (g->throbber_index) {
        case 0:
                image = &throbber1;
                g->throbber_index = 1;
                break;

        case 1:
                image = &throbber2;
                g->throbber_index = 2;
                break;

        case 2:
                image = &throbber3;
                g->throbber_index = 3;
                break;

        case 3:
                image = &throbber4;
                g->throbber_index = 4;
                break;

        case 4:
                image = &throbber5;
                g->throbber_index = 5;
                break;

        case 5:
                image = &throbber6;
                g->throbber_index = 6;
                break;

        case 6:
                image = &throbber7;
                g->throbber_index = 7;
                break;

        case 7:
                image = &throbber8;
                g->throbber_index = 8;
                break;

        case 8:
                image = &throbber0;
                g->throbber_index = 0;
                break;

        default:
                return;
        }

        if (g->throbber_index >= 0) {
                fbtk_set_bitmap(g->throbber, image);
                schedule(10, throbber_advance, g);
        }
}

void gui_window_start_throbber(struct gui_window *g)
{
        g->throbber_index = 0;
        schedule(10, throbber_advance, g);
}

void gui_window_stop_throbber(struct gui_window *gw)
{
        gw->throbber_index = -1;
        fbtk_set_bitmap(gw->throbber, &throbber0);

        fb_update_back_forward(gw);

}

void gui_window_place_caret(struct gui_window *g, int x, int y, int height)
{
        gui_window_remove_caret(g);

        // doesn't work right
        // either we'll get carets everywhere or none at all
        bbox_t savedata; 
        fbtk_draw_begin(g->browser, &savedata);
        int scroll_x, scroll_y; 
        gui_window_get_scroll(g, &scroll_x, &scroll_y); 
        
        int screen_x = x + fbtk_get_x(g->browser) - scroll_x;
        int screen_y = y + fbtk_get_y(g->browser) - scroll_y;
        LOG(("sx=%d sy=%d h=%d", screen_x, screen_y, height));
        plot.line(screen_x, screen_y, screen_x, screen_y+height, 1, 0xFF000000, false, false); 
        fbtk_draw_end(g->browser, &savedata);
        
        g->c_x = x;
        g->c_y = y;
        g->c_h = height;
        
        bbox_t rbox; 
        rbox.x0 = screen_x-1; 
        rbox.y0 = screen_y; 
        rbox.x1 = screen_x+1;
        rbox.y1 = screen_y+height; 
        fb_os_redraw(&rbox); 
}

void gui_window_remove_caret(struct gui_window *g)
{
        if(g->c_h <= 0) return; 
        
        bbox_t savedata; 
        fbtk_draw_begin(g->browser, &savedata);
        int scroll_x, scroll_y; 
        gui_window_get_scroll(g, &scroll_x, &scroll_y); 
        
        int screen_x = g->c_x + fbtk_get_x(g->browser) - scroll_x;
        int screen_y = g->c_y + fbtk_get_y(g->browser) - scroll_y;
        LOG(("sx=%d sy=%d h=%d", screen_x, screen_y, g->c_h));
        plot.line(screen_x, screen_y, screen_x, screen_y+g->c_h, 1, 0xFFFFFFFF, false, false); 
        fbtk_draw_end(g->browser, &savedata);
        
        bbox_t rbox; 
        rbox.x0 = screen_x-1; 
        rbox.y0 = screen_y; 
        rbox.x1 = screen_x+1;
        rbox.y1 = screen_y+g->c_h; 
        
        g->c_x = 0;
        g->c_y = 0;
        g->c_h = 0;
        
        fb_os_redraw(&rbox); 
}

void gui_window_new_content(struct gui_window *g)
{
}

bool gui_window_scroll_start(struct gui_window *g)
{
        return true;
}

bool gui_window_box_scroll_start(struct gui_window *g,
                int x0, int y0, int x1, int y1)
{
        return true;
}

bool gui_window_frame_resize_start(struct gui_window *g)
{
        LOG(("resize frame\n"));
        return true;
}

void gui_window_save_as_link(struct gui_window *g, struct content *c)
{
}

void gui_window_set_scale(struct gui_window *g, float scale)
{
        LOG(("set scale\n"));
}

struct gui_download_window *gui_download_window_create(const char *url,
                const char *mime_type, struct fetch *fetch,
                unsigned int total_size, struct gui_window *gui)
{
        return NULL;
}

void gui_download_window_data(struct gui_download_window *dw, const char *data,
                unsigned int size)
{
}

void gui_download_window_error(struct gui_download_window *dw,
                const char *error_msg)
{
}

void gui_download_window_done(struct gui_download_window *dw)
{
}

void gui_drag_save_object(gui_save_type type, struct content *c,
                struct gui_window *g)
{
}

void gui_drag_save_selection(struct selection *s, struct gui_window *g)
{
}

void gui_start_selection(struct gui_window *g)
{
}

void gui_paste_from_clipboard(struct gui_window *g, int x, int y)
{
}

bool gui_empty_clipboard(void)
{
        return false;
}

bool gui_add_to_clipboard(const char *text, size_t length, bool space)
{
        return false;
}

bool gui_commit_clipboard(void)
{
        return false;
}

bool gui_copy_to_clipboard(struct selection *s)
{
        return false;
}

void gui_create_form_select_menu(struct browser_window *bw,
                struct form_control *control)
{
}

void gui_launch_url(const char *url)
{
}

bool gui_search_term_highlighted(struct gui_window *g,
                unsigned start_offset, unsigned end_offset,
                unsigned *start_idx, unsigned *end_idx)
{
        return false;
}



void gui_cert_verify(struct browser_window *bw, struct content *c,
                const struct ssl_cert_info *certs, unsigned long num)
{
}

/*
 * Local Variables:
 * c-basic-offset:8
 * End:
 */
