
#define FB_SCROLL_COLOUR 0xFF888888
#define FB_FRAME_COLOUR 0xFFDDDDDD
#define FB_COLOUR_BLACK 0xFF000000
#define FB_COLOUR_WHITE 0xFFFFFFFF

enum {
        FB_KEY_HOME = 0x1000,
        FB_KEY_MENU,
        FB_KEY_BACK,
        FB_KEY_NAVMODE
}; 

typedef struct fbtk_widget_s fbtk_widget_t;

/* user widget callback */
typedef int (*fbtk_user_t)(fbtk_widget_t *widget, void *pw);

/* input callback */
typedef int (*fbtk_input_t)(fbtk_widget_t *widget, int value, void *pw);

/* mouse click callback */
typedef int (*fbtk_mouseclick_t)(fbtk_widget_t *widget, browser_mouse_state st, int x, int y, void *pw);

/* mouse move callback */
typedef int (*fbtk_move_t)(fbtk_widget_t *widget, int x, int y, void *pw);

/* redraw function */
typedef int (*fbtk_redraw_t)(fbtk_widget_t *widget, void *pw);

/* enter pressed on writable icon */
typedef int (*fbtk_enter_t)(void *pw, char *text);

/* focus gained / lost */
typedef int (*fbtk_focus_t)(fbtk_widget_t *widget, bool gain_focus, void *pw);

/* Widget creation */

/** Initialise widget toolkit.
 *
 * Initialises widget toolkit and creates root window against a framebuffer.
 *
 * @param fb The underlying framebuffer.
 * @return The root widget handle.
 */
fbtk_widget_t *fbtk_init(framebuffer_t *fb);

/** Create a window widget.
 *
 * @param parent The parent window or the root widget for a top level window.
 * @param x The x location relative to the parent window.
 * @param y the y location relative to the parent window.
 * @param width The width of the window. 0 indicates parents width should be
 *              used. Negative value indicates parents width less the value
 *              should be used. The width is limited to lie within the parent
 *              window.
 * @param height The height of the window limited in a similar way to the
 *               /a width.
 * @param c The background colour.
 * @return new window widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_window(fbtk_widget_t *parent, int x, int y, int width, int height);

/** Create a text widget.
 *
 * @param window The window to add the text widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_text(fbtk_widget_t *window, int x, int y, int width, int height, colour bg, colour fg, bool outline);

/** Create a bitmap widget.
 *
 * Create a widget which shows a bitmap.
 *
 * @param window The window to add the bitmap widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_bitmap(fbtk_widget_t *window, int x, int y, colour c,struct bitmap *image);

/** Create a filled rectangle
 *
 * Create a widget which is a filled rectangle, usually used for backgrounds.
 *
 * @param window The window to add the filled area widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *
fbtk_create_fill(fbtk_widget_t *window, int x, int y, int width, int height, colour c);

/** Create a horizontal scroll widget
 *
 * Create a horizontal scroll widget.
 *
 * @param window The window to add the filled area widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *
fbtk_create_hscroll(fbtk_widget_t *window, int x, int y, int width, int height, colour fg, colour bg);

/** Create a vertical scroll widget
 *
 * Create a vertical scroll widget.
 *
 * @param window The window to add the filled area widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *
fbtk_create_vscroll(fbtk_widget_t *window, int x, int y, int width, int height, colour fg, colour bg);

/** Create a user widget.
 *
 * Create a widget which is to be handled entirely by the calling application.
 *
 * @param window The window to add the user widget to.
 * @param pw The private pointer which can be read using ::fbtk_get_pw
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_user(fbtk_widget_t *window, int x, int y, int width, int height, void *pw);


/** Create a button widget.
 *
 * Helper function which creates a bitmap widget and associate a handler for
 * when it is clicked.
 *
 * @param window The window to add the button widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_button(fbtk_widget_t *window, int x, int y, colour c, struct bitmap *image, fbtk_mouseclick_t click, void *pw);

/** Create a writable text widget.
 *
 * Helper function which creates a text widget and configures an input handler
 * to create a writable text field. This call is equivalent to calling
 * ::fbtk_create_text followed by ::fbtk_writable_text
 *
 * @param window The window to add the text widget to.
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_create_writable_text(fbtk_widget_t *window, int x, int y, int width, int height, colour bg, colour fg, bool outline, fbtk_enter_t enter, void *pw);

/** Pop up a menu. 
 * @param parent The parent window upon which the menu is popped up. When the menu is closed focus will be restored to its parent window.
 * @param options The menu option texts, separated by newlines. 
 * @param select Callback when a menu option is selected. 
 * @return new widget handle or NULL on error.
 */
fbtk_widget_t *fbtk_popup_menu(fbtk_widget_t *parent, fbtk_widget_t *menu_parent, int x, int y, colour bg, colour fg, 
                               const char *options, fbtk_input_t select, void *pw);
/** Return to parent menu, or close menu if this was the top-level menu. 
 * @param submenu Open submenu to be closed. 
 */
void fbtk_menu_return_to_parent(fbtk_widget_t *submenu);
/** Close all open menus. 
 * @param submenu Open sub-menu (don't use the main menu) to be closed. 
 */
void fbtk_menu_close(fbtk_widget_t *submenu); 

/* Widget Destruction */

/** Destroy and free a widget and all its children.
 *
 * @param widget The widget to destroy.
 * @return 0 on success or -1 on error.
 */
int fbtk_destroy_widget(fbtk_widget_t *widget);

/* Widget information */

int fbtk_get_y(fbtk_widget_t *widget);
int fbtk_get_x(fbtk_widget_t *widget);
int fbtk_get_width(fbtk_widget_t *widget);
int fbtk_get_height(fbtk_widget_t *widget);
void *fbtk_get_userpw(fbtk_widget_t *widget);
fbtk_widget_t *fbtk_get_root_widget(fbtk_widget_t *widget);
bool fbtk_has_focus(fbtk_widget_t* widget);
fbtk_widget_t *fbtk_get_focused_widget(fbtk_widget_t* widget);
bool fbtk_is_editable(fbtk_widget_t *widget); // returns true if this is a text-writing widget
const char *fbtk_menu_get_text(fbtk_widget_t *menu, int sel); // sel counts from 1
/** Returns widget-relative coordinates of the bounding box of a menu entry. 
  * @param entry Entry number, starting at 1 for the first entry. 
  */
bool fbtk_menu_get_bbox(fbtk_widget_t *widget, int entry, bbox_t *bbox);
                
/* Set widget properties */
// addition: draw in widget
void fbtk_draw_begin(fbtk_widget_t *widget, bbox_t *savedata);
void fbtk_draw_end(fbtk_widget_t *widget, bbox_t *savedata);

// addition: programmatically change window focus
void fbtk_set_focus(fbtk_widget_t *widget);

void fbtk_set_text(fbtk_widget_t *widget, const char *text);
void fbtk_set_bitmap(fbtk_widget_t *widget, struct bitmap *image);
void fbtk_set_scroll(fbtk_widget_t *widget, int pct);
void fbtk_set_scroll_pos(fbtk_widget_t *widget, int pos);
void fbtk_set_pos_and_size(fbtk_widget_t *widget, int x, int y, int width, int height);
fbtk_redraw_t fbtk_set_handler_redraw(fbtk_widget_t *widget, fbtk_redraw_t input, void *pw);
fbtk_input_t fbtk_set_handler_input(fbtk_widget_t *widget, fbtk_input_t input, void *pw);
fbtk_mouseclick_t fbtk_set_handler_click(fbtk_widget_t *widget, fbtk_mouseclick_t click, void *pw);
fbtk_move_t fbtk_set_handler_move(fbtk_widget_t *widget, fbtk_move_t move, void *pw);
fbtk_focus_t fbtk_set_handler_focus(fbtk_widget_t *widget, fbtk_focus_t focus, void *pw);

/** Alter a text widget to be writable.
 */
void fbtk_writable_text(fbtk_widget_t *widget, fbtk_enter_t enter, void *pw);


/* General routines */

bool fbtk_clip_rect(const bbox_t * restrict clip, bbox_t * restrict box);

/** Pointer movement.
 *
 * Pointer has been moved.
 *
 * @param widget any tookit widget.
 * @parm x movement in horizontal plane.
 * @parm y movement in vertical plane.
 * @parm relative Wether the /a x and /a y should be considered relative to
 *                current pointer position.
 */
void fbtk_move_pointer(fbtk_widget_t *widget, int x, int y, bool relative);

/** Mouse has been clicked
 */
void fbtk_click(fbtk_widget_t *widget, browser_mouse_state st);

/** Input has been recived
 */
void fbtk_input(fbtk_widget_t *widget, uint32_t ucs4);

/** Indicate a widget has to be redrawn
 */
void fbtk_request_redraw(fbtk_widget_t *widget);

/** Cause a redraw to happen.
 */
int fbtk_redraw(fbtk_widget_t *widget);

int fbtk_count_children(fbtk_widget_t *widget);












